; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; ---------------------------------------------------------------------------
;                .::** ISRs and routines called by them **::.
; ---------------------------------------------------------------------------

; **
; VBlankFunc:  Called every VBlanking period (59.73 Hz).
;
;              - Updates the OAM from the Local OAM, at the same time
;                updating all 8 object palettes.
;              - Copies the bitmap patterns of MOs that have been queued
;                in MOVELIST (taking into account the various attribute
;                bits set in "MO").
;              - Copies cards that have been queued in COPYLIST to VRAM
;                (up to 12 at a time).
;
; Notes:       VBlank lasts for only 2,280 cycles in 2x CPU mode.
;
;              No more than 950 cycles of "interrupt lag" would be
;              acceptable in order for VBlankFunc to execute properly.
;

VBlankFunc:

        ld              [VF_M2 + 1], sp ; save sp at interrupt - 2

        ld              sp, vbfRegs + 4 * 2

        push            af
        push            bc
        push            de
        push            hl              ; save registers to vbfRegs

        ; ^ 24 cycles, 10 bytes

        ; -------------------------------------------------------------------
        ;                 Update Object Color Palettes
        ; -------------------------------------------------------------------

        ld              hl, Page_Ready

        ld              a, [hl]
        dec             a               ; is PageReady = 1 ?
        jp              nz, PageReady_Not

        ; ^ 15/14 cycles, 12 bytes

        ld              [hl+], a        ; set PageReady = 0

;        IF              COUNT_FRAMES == 1
;
;        inc             [hl]            ; inc CurFrames
;
;        ENDC

        ld              a, 5            ; a = #/WRAM bank where LocalOAM is
        ldh             [rSVBK], a      ; switch to WRAM bank #5

        ; prepare registers for CopyToOAM (to save space in High RAM)

        ld              bc, 8 << 8 + rOCPS & $FF  ; use b to advance palette
                                                  ; c -> rOCPS
        ld              hl, rOCPD       ; hl -> Obj. Color Palette Data Reg.
        ld              sp, MO_Colors   ; sp -> MO_Colors (at palette #0)

        ld              a, $C2
        ld              [c], a          ; auto-advance, palette #0, color #1

        REPT            3               ; update colors #1-#3 of palette #0

        pop             de              ; de <- RGB color value of entry

        ld              [hl], e         ; store LOW byte of RGB color value
        ld              [hl], d         ; store HIGH byte of RGB color value

        ENDR

        add             b
        ld              [c], a          ; auto-advance, palette #1, color #1

        REPT            3               ; update colors #1-#3 of palette #1

        pop             de              ; de <- RGB color value of entry

        ld              [hl], e         ; store LOW byte of RGB color value
        ld              [hl], d         ; store HIGH byte of RGB color value

        ENDR

        add             b

        jp              CopyToOAM       ; copy local OAM to real OAM

        ; ^ 65 cycles

CopyToOAM_Ret:

        ; finish updating palettes (only #2-#6 were updated By CopyToOAM)

        pop             af              ; a <- OCPS register "command" byte
                                        ; f <- C = 1 if vertical blockout on
                                        ;      C = 0 if vertical blockout off

        pop             de              ; de <- RGB color value of border

        jr              nc, .VBlockOut_Off

        ; overwrite one of the MO's colors with the border color so that the
        ; sprites for the vertical "block-out" column have the correct color

        ld              [c], a          ; pos @ appropriate palette and color

        ld              [hl], e         ; store LOW byte of RGB color value
        ld              [hl], d         ; store HIGH byte of RGB color value

.VBlockOut_Off

        ; update palette #0, color #0 w/ "new" border color

        ld              l, rBCPS & $FF  ; hl -> Bkg Color Palette Spec

        ld              a, $80          ; auto-advance, palette #0, color #0
        ld              [hl+], a

        ld              [hl], e         ; store LOW byte of RGB color value
        ld              [hl], d         ; store HIGH byte of RGB color value

        ld              sp, WINPAL_0 + 2 * 1

        push            de              ; palette #0, color #0 <- BrdrColRGB

        ; ^ 24/30 cycles

        ; -------------------------------------------------------------------
        ;              Update Background Color Palettes #2-#7
        ; -------------------------------------------------------------------

        ld              a, 3            ; a = #/WRAM bank where PALTAB is
        ldh             [rSVBK], a      ; switch to WRAM bank #3

        ld              sp, PALTAB_0 + 2 * 8  ; sp -> PALTAB @ palette #2
        dec             l                     ; hl -> Bkg Color Palette Spec

        ld              a, $90          ; auto-advance, palette #2, color #0
        ld              [hl+], a

        REPT            6 * 4           ; update palettes #2 through #7

        pop             de              ; de <- RGB color value

        ld              [hl], e         ; store LOW byte of RGB color value
        ld              [hl], d         ; store HIGH byte of RGB color value

        ENDR

        ; ^ 5 + 176 cycles

        ; -------------------------------------------------------------------
        ;                 Update Background Display Data
        ; -------------------------------------------------------------------

        inc             a               ; a = $91 (bit #0 set)
        ldh             [rVBK], a       ; switch to bank 1 / tile attributes

        ld              l, rHDMA1 & $FF ; hl -> rHDMA1 (Source Address MSB)
        xor             a

        ld              [hl], BDD_TileAttr >> 8  ; HDMA1 <- MSB of SrcAddress
        inc             l
        ld              [hl+], a                 ; HDMA2 <- LSB of SrcAddress

        ld              [hl], _SCRN0 >> 8        ; HDMA3 <- MSB of DstAddress
        inc             l
        ld              [hl+], a                 ; HDMA4 <- LSB of DstAddress

        ld              [hl], ( 12 * 32 ) >> 4 - 1  ; initiate GDMA transfer

        ; ^ 22 + 384 = 406 cycles

        ldh             [rVBK], a       ; switch to bank 0 / tile indices

        ld              l, rHDMA1 & $FF

        ld              [hl], BDD_TileNdx >> 8  ; HDMA1 <- MSB of SrcAddr
        inc             l
        ld              [hl+], a                ; HDMA2 <- LSB of SrcAddr

        ld              [hl], _SCRN0 >> 8       ; HDMA3 <- MSB of DstAddr
        inc             l
        ld              [hl+], a                ; HDMA4 <- LSB of DstAddr

        ld              [hl], ( 12 * 32 ) >> 4 - 1  ; initiate GDMA transfer

        ; ^ 20 + 384 = 404 cycles

;PageReady_Not:

        ; ^ 1100/1108 cycles

        ; -------------------------------------------------------------------
        ;                  Update bitmap patterns in VRAM
        ; -------------------------------------------------------------------

        ; NOTE:   There are approximately 1172-1180 cycles left at this
        ;         point before V-Blank period ends (LY = 149).

        ld              hl, rROMB0 & $FF00 + BANK(MISC)
        ld              [hl], l         ; select bank w/ CopyToVRAM's code

        jp              MO_CopyToVRAM   ; copy MO & card bitmaps to VRAM

        ; ^ 8 cycles

CopyToVRAM_Ret:

        ld              a, 3
        ldh             [rSVBK], a      ; switch to WRAM bank #3 (CurToPer)

        ; copy palettes #0-#1 of PALTAB_0 to WINPAL_1

        ld              sp, PALTAB_0
        ld              hl, WINPAL_1

        REPT            2 * 4 - 1       ; copy palettes #0 through #1

        pop             de              ; de <- RGB color value

        ld              a, e
        ld              [hl+], a        ; store LOW byte of RGB color value
        ld              a, d
        ld              [hl+], a        ; store HIGH byte of RGB color value

        ENDR

        pop             de              ; de <- RGB color value

        ld              a, e
        ld              [hl+], a        ; store LOW byte of RGB color value
        ld              [hl], d         ; store HIGH byte of RGB color value

        ; ^ 77 cycles

PageReady_Not:

        ld              hl, rROMB0 & $FF00 + BANK(AY8914_Update)
        ld              [hl], l         ; select ROM bank w/ AY8914_Update

        IF              COUNT_FRAMES == 1

        ld              bc, 60 << 8 + 16

        ld              hl, TotFrames

        inc             [hl]            ; inc TotFrames

        ld              a, [hl-]        ; a <- TotFrames

        cp              b               ; is TotFrames = 60 ?
        jr              c, .NotReady    ; no, so not ready to update display

        ld              a, [hl]         ; a <- CurFrames (0-60)

        cp              b               ; is CurFrames = 60 ? (set CF)

        ld              e, a
        ld              d, CurToPer >> 8  ; de -> CurToPer[ CurFrames ]

        ld              a, 0            ; preserve CF
        ld              [hl+], a        ; set CurFrames = 0
        ld              [hl], a         ; set TotFrames = 0

        ld              hl, _SCRN1 + ( 32 * 5 ) + 15

        jr              c, .GetPer      ; CurFrames < 60, so don't print '1'

        ld              a, $11          ; a = ASCII '1'
.GetPer
        ld              [hl+], a        ; print space (or ASCII '1')

        ld              a, [de]         ; a <- Percentage (in BCD)

        and             $F0             ; isolate HIGH digit in bits 4-7
        swap            a               ; put it in bits 0-3
        add             c               ; add tile base (for ASCII numerics)

        ld              [hl+], a        ; store out TileNdx (of HIGH digit)

        ld              a, [de]         ; a <- Percentage (in BCD)

        and             $0F             ; isolate LOW digit in bits 4-7
        add             c               ; add tile base (for ASCII numerics)

        ld              [hl+], a        ; store out TileNdx (of LOW digit)
        ld              [hl], 5         ; store out ASCII '%'

.NotReady

        ENDC

        ; if the LocalOAM is now current, update the rSCX & rSCY registers

        ld              hl, STIC_Analyzed

        ld              a, [hl]         ; a <- STIC_Analyzed
        dec             a               ; did STIC_Analyze finish its job ?
        jr              nz, .NotCurrent ; no, so LocalOAM is not up to date

        ld              [hl+], a        ; clear STIC_Analyzed

        ld              a, [hl+]        ; a <- rSCY_New
        ldh             [rSCY], a       ; update Scroll Y register
        ld              a, [hl+]        ; a <- rSCX_New
        ldh             [rSCX], a       ; update Scroll X register

        ld              [hl],  a        ; update rSCX_Old (for HBlankFunc)

.NotCurrent

        ; -------------------------------------------------------------------
        ;                         Update AY8914
        ; -------------------------------------------------------------------

        ld              a, 1
        ldh             [rSVBK], a      ; switch to WRAM bank #1

        jp              AY8914_Update

AY8914_Ret:

        IF              SYNC_DISPLAY == 0

        ld              hl, VBlank_Ctr
        inc             [hl]

        ENDC

        ldh             a, [cSVBK]
        ldh             [rSVBK], a      ; restore RAM bank

        jp              VBlankFixUp


