; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; **
; MemCopy
;
; Parameters:  bc = ByteCount
;              de = Destination Address
;              hl = Source Address
; Modifies:    a  = [Source Address + ByteCount - 1]
;              bc = 0
;              de = Destination Address + ByteCount
;              hl = Source Address + ByteCount
;

MemCopy:
        inc             b
        inc             c
        jr              .Skip
.Loop
        ld              a, [hl+]
        ld              [de], a
        inc             de
.Skip
        dec             c
        jr              nz, .Loop
        dec             b
        jr              nz, .Loop

        ret

; **
; MemFill
;
; Parameters:  a  = value to write
;              bc = ByteCount
;              hl = Destination Address
; Modifies:    bc = 0
;              hl = Destination Address + ByteCount
;

MemFill:
        inc             b
        inc             c
        jr              .Skip
.Loop
        ld              [hl+], a
.Skip
        dec             c
        jr              nz, .Loop
        dec             b
        jr              nz, .Loop

        ret

; **
; CPU_SetSngSpeed:  Sets the GBC to run in single speed CPU mode.
;
; Parameters:       None.
;
; Modifies:         a, hl
;
; Notes:            If GBC is not already in 1x CPU mode, interrupts
;                   will be enabled when routine returns to caller.
;

CPU_SetSngSpeed:

        ; make sure not in 1x mode already, or we'll toggle it to 2x mode

        ldh             a, [rKEY1]      ; a <- Select CPU Speed
        rlca                            ; in 1x CPU mode (bit #7 reset) ?
        ret             nc              ; yes, so leave it 1x CPU mode

        jr              CPU_ToggleSpeed ; toggle from double to single speed

; **
; CPU_SetDblSpeed:  Sets the GBC to run in double speed CPU mode.
;
; Parameters:       None.
;
; Modifies:         a, hl
;
; Notes:            If GBC is not already in 2x CPU mode, interrupts
;                   will be enabled when routine returns to caller.
;

CPU_SetDblSpeed:

        ; make sure not in 2x mode already, or we'll toggle it to 1x mode

        ldh             a, [rKEY1]      ; a <- Select CPU Speed
        rlca                            ; in 2x CPU mode (bit #7 set) ?
        ret             c               ; yes, so leave it 2x CPU mode

CPU_ToggleSpeed:

        di

        ld              hl, rIE         ; hl -> Interrupt Enable ($FFFF)
        ld              a, [hl]         ; a <- interrupt enable bits

        ld              [hl], 0         ; disable all interrupts
        ld              l, rIF & $FF    ; hl -> Interrupt Flag ($FF0F)
        ld              [hl], 0         ; clear pending interrupts

        ld              l, rP1 & $FF         ; hl -> Joypad Info ($FF00)
        ld              [hl], P1F_4 + P1F_5  ; "de-select" P14 & P15 (reset)

        ld              l, rKEY1 & $FF  ; hl -> Select CPU Speed ($FF4D)
        ld              [hl], 1         ; select 2x CPU mode

        stop                            ; wait for button press (???)

        ld              l, h            ; hl -> Interrupt Enable ($FFFF)
        ld              [hl], a         ; restore interrupt enable bits

        reti                            ; return, enabling interrupts

; **
; WriteStr:
;
; Parameters:  a  = Attribute Byte
;              b  = Tile Base
;              de = Source Address
;              hl = Destination Address
;
; Modifies:    c, de, hl
;

WriteStr:

        ; first write the string and count up the number of characters in it

        push            af              ; save Attribute Byte

        ld              c, 0            ; set #/characters to display = 0

        jr              .Char_VRAM

.Char_Loop

        inc             c               ; inc #/characters to display
        add             b               ; add Tile Base
        ld              [hl+], a        ; store out Tile Index

.Char_VRAM

        ldh             a, [rSTAT]      ; a <- LCDC Status
        and             STATF_BUSY      ; is VRAM safe to write to ?
        jr              nz, .Char_VRAM  ; nope, so wait until it is

        ld              a, [de]         ; a <- character from string
        inc             de              ; increment string index
        or              a               ; is it a 0 ?
        jr              nz, .Char_Loop  ; yes, so we've hit the end of string

        ; now write the tile attributes and restore HL to its original value

        inc             a               ; a = 1
        ldh             [rVBK], a       ; select tile attributes

        pop             af              ; restore Attribute Byte

        push            hl              ; save Destination Address
        push            bc              ; save Tile Base and #/characters

        ld              b, a            ; b = Attribute Byte
        dec             hl              ; position HL at the last character

.Attrib_Loop

        ldh             a, [rSTAT]        ; a <- LCDC Status
        and             STATF_BUSY        ; is VRAM safe to write to ?
        jr              nz, .Attrib_Loop  ; nope, so wait until it is

        ld              a, b            ; a = Attribute Byte
        ld              [hl-], a        ; store out Attribute Byte

        dec             c               ; dec #/characters to display
        jr              nz, .Attrib_Loop

        xor             a
        ldh             [rVBK], a       ; select tile indices

        ld              a, b            ; restore Attribute Byte to A

        pop             bc              ; restore Tile Base and #/characters
        pop             hl              ; restore Destination Address

        ret

; **
; WriteHex:    Write a hexidecimal value to the background.
;
; Parameters:  A  = Attribute Byte
;              C  = #/digits to display (i.e., 1-4)
;              B  = Tile Base
;              DE = 16-bit value (to write in hex)
;              HL = Destination Address
;
; Modifies:    c  = 0
;              de = DE >> ( C * 4 )
;              hl = HL + C
;

WriteHex:

        ; first set the tile attribute and position HL at the last digit

        push            af              ; save Attribute Byte
        push            bc              ; save Tile Base & #/digits

        ld              b, a            ; save Attribute Byte in b

        ld              a, 1
        ldh             [rVBK], a       ; select tile attributes

.Attrib_Loop

        ldh             a, [rSTAT]
        and             STATF_BUSY        ; is VRAM safe to write to ?
        jr              nz, .Attrib_Loop  ; no, so wait until it is

        ld              a, b            ; restore Attribute Byte from b
        ld              [hl+], a        ; store out Attribute Byte

        dec             c               ; dec #/digits to display
        jr              nz, .Attrib_Loop

        xor             a
        ldh             [rVBK], a       ; select tile indices

        pop             bc              ; restore Tile Base & #/digits

        push            hl              ; save position in background

        dec             hl              ; position HL at last digit

        ; now convert each nybble to '1'..'9' or 'A'..'F' and store it out

.Char_Loop

        ldh             a, [rSTAT]      ; a <- LCDC Status
        and             STATF_BUSY      ; is VRAM access safe ?
        jr              nz, .Char_Loop  ; nope, so wait until it is

        ld              a, e
        and             $0F             ; isolate lowest nybble of DE
        add             $30             ; put it in '0'..'?'

        cp              $3A             ; is it in ':'..'?' ?
        jr              c, .Numeric     ; no, so in '1'..'9', which is OK

        add             $41 - $3A       ; put it in 'A'..'F'
.Numeric
        add             b               ; add TileBase
        ld              [hl-], a        ; store out TileIndex

        ld              a, e

        srl             d
        rra
        srl             d
        rra
        srl             d
        rra
        srl             d
        rra

        ld              e, a            ; de = de >> 4, a = e

        dec             c               ; dec #/digits to write
        jr              nz, .Char_Loop

        pop             hl              ; restore position in background
        pop             af              ; restore Attribute Byte

        ret

; **
; RepChar:     Write the same character to the background X times.
;
; Parameters:  A  = Attribute Byte
;              C  = Character to Repeat
;              B  = Tile Base
;              DE = #/times to repeat character
;              HL = Destination Address
;
; Modifies:    de = 0
;              hl = HL + DE
;

RepChar:

        ; first, write out tile attributes & position HL @ the final position

        push            af              ; save Attribute Byte
        push            bc              ; save Tile Base, Char. to Repeat
        push            de              ; save #/times to repeat character

        ld              b, a            ; save Attribute Byte in b

        ld              a, 1
        ldh             [rVBK], a       ; select tile attributes

        inc             e
        inc             d               ; prepare DE for loop

        jr              .Attrib_Skip

.Attrib_Loop

        ldh             a, [rSTAT]        ; a <- LCDC Status
        and             STATF_BUSY        ; is it safe to write to VRAM ?
        jr              nz, .Attrib_Loop  ; no, so wait until it is

        ld              a, b
        ld              [hl+], a        ; store out Attribute Byte
.Attrib_Skip
        dec             e
        jr              nz, .Attrib_Loop
        dec             d
        jr              nz, .Attrib_Loop

        pop             de              ; restore #/times to repeat character
        pop             bc              ; restore Tile Base, Char. to Repeat

        push            hl              ; save HL at its final position

        dec             hl              ; position HL at the last character

        ; now write the character (in C) DE times to the background

        xor             a
        ldh             [rVBK], a       ; select tile indices

        inc             e
        inc             d               ; prepare DE for loop

        jr              .Char_Skip

.Char_Loop

        ldh             a, [rSTAT]      ; a <- LCDC Status
        and             STATF_BUSY      ; is it safe to write to VRAM ?
        jr              nz, .Char_Loop  ; no, so wait until it is

        ld              a, c            ; a = Character to Repeat
        add             b               ; add Tile Base
        ld              [hl-], a        ; store out Tile Index
.Char_Skip
        dec             e
        jr              nz, .Char_Loop
        dec             d
        jr              nz, .Char_Loop

        pop             hl              ; save HL at its final position
        pop             af              ; restore Attribute Byte

        ret

