; **
; Intellivision for GameBoy Color v1.1
; Copyright(C) 2000,03 By Carl Mueller, Jr.
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

; ---------------------------------------------------------------------------
;                  .::** CP1610 Emulation Routines **::.
; ---------------------------------------------------------------------------

; **
; Intv_Reset:  Initialize the CP1610 to its "power-on" state by:
;
;              - Setting Status and SDBD to 0.
;              - Setting [R7] to the address of the EXEC's 1st program word.
;              - Switching to the EXEC bank and setting cROMB0 and R7_Conv
;                appropriately.
;              - Setting RIGHT and LEFT hand controller values to $FF.
;              - Processing instructions until a "break point" is reached
;                (see ifBcond_BREAK and ifDECR_R7).
;
; Notes:       Intv_Reset must be run prior to calling Process.
;
; Parameters:  None.
;
; Modifies:    All registers.
;

Intv_Reset:

        xor             a
        ldh             [STIC_Display_Hi], a  ; turn STIC's display off
        ld              [STIC_Display], a
        ld              [SDBD], a

        ld              bc, $2000 + cROMB0 & $FF  ; bc -> rROMB0
                                                  ;  c -> cROMB0

        ld              a, BANK(EXEC)
        ld              [c], a                         ; [cROMB0]
        ld              [bc], a                        ; [rROMB0]
        ld              a, ( $1000 - EXEC >> 1 ) >> 8
        ldh             [R7_Conv], a                   ; [R7_Conv]

        ld              hl, SCRATCHPAD_RAM + $FF * 2
        ld              a, $FF
        ld              [hl-], a        ; set left-hand controller
        dec             l
        ld              [hl], a         ; set right-hand controller

        ld              [SP_0], sp      ; save stack pointer in SP_0
        di                              ; so V/HBlankFunc can't interrupt
        ld              sp, EXEC        ; sp -> EXEC's 1st program word

        reti                            ; return, enabling interrupts

; **
; Process:    Process instructions until a "break point" is reached (see
;             ifBcond_BREAK, ifDECR_R7, and ifHLT), first interrupting the
;             current instruction (referenced by [R7]), pushing its address
;             on the stack, and running the ISR at $1004.
;
; Notes:      Intv_Reset must be run prior to calling Process.
;
; Parameters: None.
;
; Modifies:   All registers.
;

Process:

        ld              a, [Status]     ; a <- ..1.SZOC (Status)

        ld              c, cROMB0 & $FF
        ld              b, a            ; bc = rROMB0 / Status / cROMB0

        ld              [SP_0], sp      ; save stack pointer in SP_0

        di                              ; don't let VBlankFunc interrupt

        ld              sp, R7
        pop             hl
        ld              sp, hl          ; sp <- [R7] (Real)

        ld              d, h
        ld              a, l            ; da = [R7] (Real)

        srl             d
        rra                             ; da = [R7] >> 1 (Real)

        ld              hl, R6          ; hl -> R6

        inc             [hl]            ; [R6] <- [R6] + 1

        ld              l, [hl]

        ld              [hl-], a        ; store out LOW of [R7] (CP1610)
        inc             h

        ldh             a, [R7_Conv]
        add             d
        ld              [hl], a         ; store out HIGH byte of [R7]

        ld              a, $10
        ld              e, $04          ; JumpAddress = $1004 (CP1610)

        jp              Jump_Op         ; conv. addr, switch banks if needed

; ---------------------------------------------------------------------------
;                .::** CP1610 Implementing Function Code **::.
; ---------------------------------------------------------------------------

ifHLT::                                 ; HaLT

        ; OBJ_00:       ifDECR_R7_1

ifSDBD::                                ; Set Double Byte Data (mode)

        ; OBJ_00:       ifSDBD

        ld              a, c            ; = ld a, $D3 (non-zero value)
        ld              [SDBD], a       ; turn Double Byte Data mode ON

        RETURN          0               ; return, enabling interrupts

ifEIS::                                 ; Enable Interrupt System

        ; OBJ_00:       ifNOP

ifDIS::                                 ; Disable Interrupt System

        ; OBJ_00:       ifNOP

ifCLRC::                                ; CLeaR Carry

        ; OBJ_00:       ifCLRC

        res             0, b            ; clear Carry (bit #0)

        RETURN          0               ; return, enabling interrupts

ifSETC::                                ; SET Carry

        ; OBJ_00:       ifSETC

        set             0, b            ; set Carry (bit #0)

        RETURN          0               ; return, enabling interrupts

; ---------------------------------------------------------------------------
;  .::** Object definitions for EXEC HELPER routines -- see EHJSR.ASM **::.
; ---------------------------------------------------------------------------

ifJSR_1629::

        ; OBJ_00:       EH_1629
        ; OBJ_01:       R0
        ; OBJ_02:       $0080

ifJSR_1668::

        ; OBJ_00:       EH_1668
        ; OBJ_01:       R0
        ; OBJ_02:       $FF00

ifJSR_167D::

        ; OBJ_00:       EH_167D
        ; OBJ_01:       R0
        ; OBJ_02:       SYSTEM_RAM + $015E * 2

ifJSR_1757::

        ; OBJ_00:       EH_1757
        ; OBJ_01:       R0 + 1
        ; OBJ_02:       $0100

ifJSR_1766::

        ; OBJ_00:       EH_1766
        ; OBJ_01:       R1
        ; OBJ_02:       $071D

ifJSR_176C::

        ; OBJ_00:       EH_176C
        ; OBJ_01:       R0
        ; OBJ_02:       R1

ifJSR_1A2F::

        ; OBJ_00:       EH_1A2F
        ; OBJ_01:       R0
        ; OBJ_02:       $2000 + cROMB0 .AND. $00FF

; ---------------------------------------------------------------------------
;         .::** Code for JSR RB, $AABB (without bank switching) **::.
; ---------------------------------------------------------------------------

ifJ::                                   ; Jump

        ; OBJ_00:       ifB
        ; OBJ_01:       XX .SHL. 8 + YY

ifJSR_R4::                              ; Jump SubRoutine, using R4

        ; OBJ_00:       ifJSR_R4
        ; OBJ_01:       PP .SHL. 8 + QQ
        ; OBJ_02:       XX .SHL. 8 + YY

        ld              l, R4 & $FF

        jr              ifJSR_R5_1

ifJSR_R5::                              ; Jump SubRoutine, using R5

        ; OBJ_00:       ifJSR_R5
        ; OBJ_01:       PP .SHL. 8 + QQ
        ; OBJ_02:       XX .SHL. 8 + YY

        ld              l, R5 & $FF

ifJSR_R5_1:                             ; entry point for ifJSR_R4 & ifJSR_R6

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RB

        pop             de              ; de <- RetAddr (CP1610)

        ld              [hl], e
        inc             l
        ld              [hl], d         ; [RB] <- RetAddr

        pop             hl
        ld              sp, hl          ; sp <- JumpAddr (Real)

        RETURN          0               ; return, enabling interrupts

        ; ^ 21 cycles, 11 bytes

;ifJSR_R6::                              ; Jump SubRoutine, using R6
;
;        ; OBJ_00:       ifJSR_R6
;        ; OBJ_01:       PP .SHL. 8 + QQ
;        ; OBJ_02:       XX .SHL. 8 + YY
;
;        ld              l, R6 & $FF
;
;        jr              ifJSR_R5_1

; ---------------------------------------------------------------------------
;          .::** Code for JSR RB, $AABB (with bank switching) **::.
; ---------------------------------------------------------------------------

ifJ_B::                                 ; Jump w/ Bank (change)

        ; OBJ_00:       ifJ_B
        ; OBJ_01:       AA .SHL. 8 + BB
        ; OBJ_02:       $4D43

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> AddrHiToROMB0

        jr              ifJSR_R5_B_2

ifJSR_R4_B::                            ; Jump SubRoutine, using R4

        ; OBJ_00:       ifJSR_R4_B
        ; OBJ_01:       PP .SHL. 8 + QQ
        ; OBJ_02:       AA .SHL. 8 + BB

        ld              l, R4 & $FF

        jr              ifJSR_R5_B_1

ifJSR_R5_B::                            ; Jump SubRoutine, using R5

        ; OBJ_00:       ifJSR_R5_B
        ; OBJ_01:       PP .SHL. 8 + QQ
        ; OBJ_02:       AA .SHL. 8 + BB

        ld              l, R5 & $FF

ifJSR_R5_B_1:

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RB

        pop             de              ; de <- RetAddr (CP1610)

        ld              [hl], e
        inc             l
        ld              [hl], d         ; [RB] <- RetAddr

ifJSR_R5_B_2:                           ; entry point for ifJ_B

        pop             de              ; de <- JumpAddr (CP1610)

        ld              l, d            ; l = HIGH of JumpAddr (CP1610)

        ld              a, [hl]         ; a <- new ROM bank #
        ld              [c], a          ; note new ROM bank # for ISRs
        ld              [bc], a         ; switch to new ROM bank

        inc             h
        ld              h, [hl]         ; a <- HIGH of JumpAddr (Real) >> 1

        ld              a, l            ; a = HIGH of JumpAddr (CP1610)
        sub             h
        ldh             [R7_Conv], a

        ld              l, e
        add             hl, hl

        ld              sp, hl          ; sp = JumpAddr (Real)

        RETURN          0               ; return, enabling interrupts

        ; ^ 35 + 4 = 39 cycles, 23 bytes

; ---------------------------------------------------------------------------
;              .::** Code for JSR RB, $AABB (to the EXEC) **::.
; ---------------------------------------------------------------------------

ifJ_EXEC::                              ; Jump w/ Bank (change)

        ; OBJ_00:       ifJ_EXEC
        ; OBJ_01:       XX .SHL. 8 + YY
        ; OBJ_02:       $4D43

        di                              ; don't allow VBlankFunc to interrupt

        jr              ifJSR_EXEC_R5_2

ifJSR_1F35::

        ; OBJ_00:       ifJSR_1F35
        ; OBJ_01:       PP .SHL. 8 + QQ
        ; OBJ_02:       XX .SHL. 8 + YY

        ; disable display while background cards are loaded

        xor             a
        ldh             [STIC_Display_Hi], a

ifJSR_EXEC_R4::                         ; Jump SubRoutine, using R4

        ; OBJ_00:       ifJSR_EXEC_R4
        ; OBJ_01:       PP .SHL. 8 + QQ
        ; OBJ_02:       XX .SHL. 8 + YY

        ld              l, R4 & $FF

        jr              ifJSR_EXEC_R5_1

ifJSR_1F2F::

        ; OBJ_00:       ifJSR_1F2F
        ; OBJ_01:       PP .SHL. 8 + QQ
        ; OBJ_02:       XX .SHL. 8 + YY

        ; disable display while background cards are loaded

        xor             a
        ldh             [STIC_Display_Hi], a

ifJSR_EXEC_R5::                         ; Jump SubRoutine, using R5

        ; OBJ_00:       ifJSR_EXEC_R5
        ; OBJ_01:       PP .SHL. 8 + QQ
        ; OBJ_02:       XX .SHL. 8 + YY

        ld              l, R5 & $FF

ifJSR_EXEC_R5_1:                        ; entry point for ifJSR_EXEC_R4

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RB

        pop             de              ; de <- RetAddr (CP1610)

        ld              [hl], e
        inc             l
        ld              [hl], d         ; [RB] <- RetAddr

ifJSR_EXEC_R5_2:                        ; entry point for ifJ_EXEC

        pop             hl              ; hl <- JumpAddr (Real)

        ld              a, BANK(EXEC)   ; a <- #/ROM bank where EXEC is
        ld              [c], a          ; note new ROM bank # for ISRs
        ld              [bc], a         ; switch to new ROM bank

        ld              a, ( $1000 - EXEC >> 1 ) >> 8
        ldh             [R7_Conv], a

        ld              sp, hl          ; sp = JumpAddr (Real)

        RETURN          0               ; return, enabling interrupts

        ; ^ 28 + 4 = 32 cycles, 19 bytes

; ---------------------------------------------------------------------------
;    CP1610's register operate instructions: INCR, DECR, COMR, NEGR, ADCR
; ---------------------------------------------------------------------------

ifINCR_F::                              ; INCrement Register w/ Flags

        ; OBJ_00:       RD_GTA|ifINCR_F

        RSING_TAB       l, 5, 0, rtRD_P1  ; l <- LOW byte of RD + 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RD + 1 ( @ HIGH byte )

        ld              a, [hl-]        ; a <- HIGH byte of [RD]

        inc             [hl]            ; increment LOW byte of [RD]
        jr              z, .ChkZero
.ChkSign
        res             2, b            ; reset Zero (bit #2)

        add             a, a            ; is sign (bit #15) set ?
        jr              c, .Sign

        res             3, b            ; reset Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.Sign

        set             3, b            ; set Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.ChkZero

        inc             l               ; pos @ HIGH byte of [RD]

        inc             a
        ld              [hl], a         ; increment HIGH byte of [RD]
        jr              nz, .ChkSign

        res             3, b            ; reset Sign (bit #3)
        set             2, b            ; set Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

ifINCR::                                ; INCrement Register

        ; OBJ_00:       RD_GTA|ifINCR

        RSING_TAB       l, 6, 0, rtNone ; l <- LOW byte of RD

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RD

        inc             [hl]            ; increment LOW byte of [RD]

        ret             nz              ; *

        inc             l
        inc             [hl]            ; increment HIGH byte of [RD]

        RETURN          0               ; return, enabling interrupts

ifINCR_R7::                             ; INCrement Register [R7]

        ; NOTE:  It's assumed that the flags (Status) would never be
        ;        tested as a result of an INCR R7, thus there exists
        ;        only one version of this instruction.

        ; OBJ_00:       ifINCR_R7

        di                              ; don't allow VBlankFunc to interrupt

        pop             af              ; skip next program word

        RETURN          0               ; return, enabling interrupts

ifDECR_F::                              ; DECrement Register w/ Flags

        ; OBJ_00:       RD_GTA|ifDECR_F

        RSING_TAB       l, 6, 0, rtNone ; l <- LOW byte of RD

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RD

        ld              a, -1
        add             [hl]            ; a = LOW byte of [RD] - 1
        ld              [hl+], a        ; store out LOW byte of [RD] - 1

        jr              z, .LoZero      ; if Z = 1, see if HIGH = 0 too
        jr              nc, .HiDec      ; if C = 0, need to dec HIGH byte

        ld              a, [hl]         ; a <- HIGH byte of [RD]
.ChkSign
        add             a, a            ; is sign (bit #7) set ?
        jr              c, .Sign
.NoSign
        res             3, b            ; reset Sign (bit #3)
        res             2, b            ; reset Zero (bit #2)

        ; ^ 19 cycles vs. 19 cycles

        RETURN          0               ; return, enabling interrupts

.LoZero

        or              [hl]            ; is HIGH of [RD] = 0 too ?
        jr              nz, .ChkSign    ; no, so check sign instead

        res             3, b            ; reset Sign (bit #3)
        set             2, b            ; set Zero (bit #2)

        ; ^ 17 cycles vs. 16 cycles

        RETURN          0               ; return, enabling interrupts

.HiDec

        add             [hl]            ; a = HIGH byte of [RD] - 1
        ld              [hl+], a        ; store out HIGH byte of [RD] - 1

        add             a, a            ; is sign (bit #7) set ?
        jr              nc, .NoSign
.Sign
        set             3, b            ; set Sign (bit #3)
        res             2, b            ; reset Zero (bit #2)

        ; ^ 22 cycles vs. 23 cycles

        RETURN          0               ; return, enabling interrupts

ifDECR::                                ; DECrement Register

        ; OBJ_00:       RD_GTA|ifDECR

        RSING_TAB       l, 6, 0, rtNone ; l <- LOW byte of RD

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RD

        ld              a, -1
        add             [hl]            ; a = LOW byte of [RD] - 1
        ld              [hl+], a        ; store out LOW byte of [RD]

        ret             c               ; *

        dec             [hl]            ; decrement HIGH byte of [RD]

        RETURN          0               ; return, enabling interrupts

ifCOMR_F::                              ; COMplement Register w/ Flags

        ; OBJ_00:       RD_GTA|ifCOMR_F

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RD

        ld              a, [hl]         ; a <- LOW byte of [RD]
        cpl
        ld              [hl+], a        ; store LOW byte of [RD]'s 1's comp.

        or              a
        jr              nz, .NoZero

        sub             [hl]            ; a = HIGH byte of [RD]'s 2's comp.
        dec             a               ; convert to 1's comp.
        ld              [hl], a         ; store out HIGH of [RD]'s 1's comp.

        jr              nz, .ChkSign

        set             2, b            ; set Zero (bit #2)
.NoSign
        res             3, b            ; reset Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.NoZero

        ld              a, [hl]         ; a <- HIGH byte of [RD]
        cpl
        ld              [hl], a         ; store HIGH byte of [RD]'s 1's comp.

.ChkSign

        res             2, b            ; reset Zero (bit #2)

        add             a, a            ; is sign bit set ?
        jr              nc, .NoSign

        set             3, b            ; set Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

ifCOMR::                                ; COMplement Register

        ; OBJ_00:       RD_GTA|ifCOMR

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RD

        ld              a, [hl]         ; a <- LOW byte of [RD]
        cpl
        ld              [hl+], a        ; store LOW byte of [RD]'s 1's comp.

        ld              a, [hl]         ; a <- HIGH byte of [RD]
        cpl
        ld              [hl], a         ; store HIGH byte of [RD]'s 1's comp.

        RETURN          0               ; return, enabling interrupts

ifNEGR_F::                              ; NEGate Register w/ Flags

        ; OBJ_00:       RD_GTA|ifNEGR_F

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RD

        xor             a
        sub             [hl]            ; a = LOW byte of 0 - [RD]
        ld              [hl+], a        ; store LOW byte of [RD]'s 2's comp.

        ld              a, 0            ; a = 0 (without affecting flags)
        sbc             [hl]            ;   = HIGH byte of 0 - [RD]

        jr              nc, .NoCarry    ; if CY = 0, [RD] is/was = 0

        ld              [hl-], a        ; store out HIGH byte of 2's comp.

        add             a, a            ; is bit #7 set ?
        jr              nc, .NoSign     ; no, so no possibility of overflow

        ld              b, $29          ; set Sign and Carry

        ret             nz              ; * if ZF = 1, [RD] = $80XX

        or              [hl]            ; is LOW byte of [RD] = 0 ?

        ret             nz              ; * if ZF = 1, [RD] = $8000

        ld              b, $2B          ; set Sign, Overflow and Carry

        RETURN          0               ; return, enabling interrupts

.NoSign

        ld              b, $21          ; set Carry only

        RETURN          0               ; return, enabling interrupts

.NoCarry

        ld              b, $24          ; set Zero only

        RETURN          0               ; return, enabling interrupts

ifNEGR::                                ; NEGate Register

        ; OBJ_00:       RD_GTA|ifNEGR

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RD

        xor             a               ; a = 0
        sub             [hl]            ; a = LOW byte of 0 - [RD]
        ld              [hl+], a        ; store LOW byte of [RD]'s 2's comp.

        ld              a, 0            ; a = 0 (without affecting flags)
        sbc             [hl]            ; a = HIGH byte of 0 - [RD]
        ld              [hl], a         ; store HIGH byte of [RD]'s 2's comp.

        RETURN          0               ; return, enabling interrupts

ifADCR_F::                              ; ADd Carry to Register w/ Flags

        ; OBJ_00:       RD_GTA|ifADCR_F

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RD

        ld              a, b            ; a = ..1.SZOC (Status)
        and             $01             ;   = .......C
        jr              z, .CarryNotSet

        add             [hl]            ; a = LOW byte of [RD] + 1
        ld              [hl+], a        ; store out LOW byte of [RD] + 1

        ld              a, 0
        adc             [hl]            ; a = HIGH byte of [RD] + 1
        ld              [hl-], a        ; store out HIGH byte of [RD] + 1

        jr              c, .ZeroAndCarry  ; if CY = 1, [RD] = 0

        add             a, a            ; is bit #7 set ?
        jr              nc, .NoFlags    ; no, so no possibility of overflow

        ld              b, $28          ; set Sign bit only

        ret             nz              ; * if Z = 1, HIGH byte of [RD] = $80

        or              [hl]            ; is LOW of [RD] = 0 ?

        ret             nz              ; * if Z = 0, [RD] = $8000

        ld              b, $2A          ; set Sign and Overflow bits

        RETURN          0               ; return, enabling interrupts

.ZeroAndCarry

        ld              b, $25          ; set Zero and Carry

        RETURN          0               ; return, enabling interrupts

.CarryNotSet

        inc             l               ; pos @ HIGH byte of [RD]
        ld              a, [hl-]        ; a <- HIGH byte of [RD]

        add             a, a            ; is sign (bit #7) set ?
        jr              c, ifCMPR_DIFF_F

        or              [hl]            ; are both LOW & HIGH of RD = 0 ?
        jr              z, .ZeroOnly

.NoFlags

        ld              b, $20          ; clear all flags

        RETURN          0               ; return, enabling interrupts

.ZeroOnly

        ld              b, $24          ; set Zero only

        RETURN          0               ; return, enabling interrupts

ifCMPR_DIFF_F::                         ; CoMPare Register (Same) w/ Flags

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ld              b, $28          ; set Sign only

        RETURN          0               ; return, enabling interrupts

ifADCR::                                ; ADd Carry to Register

        ; OBJ_00:       RD_GTA|ifADCR

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        bit             0, b            ; is Carry set ?
        ret             z               ; *

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RD

        inc             [hl]            ; increment LOW byte of [RD]
        ret             nz              ; *

        inc             l
        inc             [hl]            ; increment HIGH byte of [RD]

        RETURN          0               ; return, enabling interrupts

ifADCR_R7::                             ; ADd Carry to Register [R7]

        ; NOTE:  This isn't a fully implemented instruction -- i.e.,
        ;        we don't bother to set the Sign, Zero, or Overflow
        ;        flag, and the Carry flag is technically not set
        ;        accurately.  The only reason the Carry flag is
        ;        cleared is because BEAMRIDER expects it (@ $5B45).

        ; OBJ_00:       ifADCR_R7

        bit             0, b            ; is Carry set ?
        ret             z               ; *

        di                              ; don't allow VBlankFunc to interrupt

        res             0, b            ; reset Carry (see BEAMRIDER @ $5B45)

        pop             af              ; skip next program word

        RETURN          0               ; return, enabling interrupts

; ---------------------------------------------------------------------------
;                CP1610's status instructions: GSWD, RSWD
; ---------------------------------------------------------------------------

ifGSWD::                                ; Get Status WorD

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RD_GTA|ifGSWD

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR (R0-R3)

        ld              h, c            ; hl -> RR

        ld              e, b            ; e = ..1.SZOC (Status)
        res             5, e            ;     ....SZOC
        swap            e               ;     SZOC....

        ld              a, [hl]         ; a <- LOW byte of [RR]
        and             $0F             ; allow LOW nybble to pass through
        or              e               ; set UPPER nybble to SZOC
        ld              [hl+], a

        ld              a, [hl]         ; a <- HIGH byte of [RR]
        and             $0F             ; allow LOW nybble to pass through
        or              e               ; set UPPER nybble to SZOC
        ld              [hl], a

ifNOP::                                 ; No OPeration

        ; OBJ_00:       ifNOP

        RETURN          0               ; return, enabling interrupts

ifRSWD::                                ; Restore Status WorD

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RS_GTA|ifRSWD

        RSING_TAB       l, 4, 0, rtNone ; l <- LOW byte of RR (R0-R3)

        ld              h, c            ; hl -> RS

        ld              a, [hl]         ; a <- LOW byte of [RR]
        swap            a
        and             $0F             ; a = ....SZOC
        or              $20             ;     ..1.SZOC

        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

; ---------------------------------------------------------------------------
;         .::** CP1610's (register operate) SHIFT instructions **::.
; ---------------------------------------------------------------------------

ifSWAP_1_F::                            ; SWAP 1 (once) w/ Flags

        ; OBJ_00:       RD_GTA|ifSWAP_1_F

        RSING_TAB       l, 3, 0, rtRD_P1  ; l <- LOW byte of RR + 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR ( @ HIGH byte )

        ld              a, [hl-]        ; a <- HIGH byte of [RR]
        ld              d, [hl]         ; d <- LOW byte of [RR]

        ld              [hl+], a        ; store out HIGH to LOW of [RR]
        ld              [hl], d         ; store out LOW to HIGH of [RR]

        add             a, a            ; is sign (bit #7) set ?
        jr              c, .Sign

        res             3, b            ; reset Sign (bit #3)

        or              d               ; are both LOW & HIGH bytes = 0 ?
        jr              nz, .NoZero

        set             2, b            ; set Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

.Sign
        set             3, b            ; set Sign (bit #3)
.NoZero
        res             2, b            ; reset Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

ifSWAP_1::                              ; SWAP 1 (once)

        ; OBJ_00:       RD_GTA|ifSWAP_1

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              a, [hl+]        ; a <- LOW byte of [RR]
        ld              d, [hl]         ; d <- HIGH byte of [RR]

        ld              [hl-], a        ; store out LOW byte of [RR]
        ld              [hl], d         ; store out HIGH byte of [RR]

        RETURN          0               ; return, enabling interrupts

ifSWAP_2_F::                            ; SWAP 2 (twice) w/ Flags

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RD_GTA|ifSWAP_2_F

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        ld              h, c            ; hl -> RR

        ld              a, [hl+]        ; a <- LOW byte of [RR]
        ld              [hl], a         ; store out LOW byte of [RR] to HIGH

        add             a, a            ; is sign (bit #7) set ?
        jr              c, .Sign

        res             3, b            ; reset Sign (bit #3)

        jr              nz, .NoZero

        set             2, b            ; set Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

.Sign
        set             3, b            ; set Sign (bit #3)
.NoZero
        res             2, b            ; reset Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

ifSWAP_2::                              ; SWAP 2 (twice)

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RD_GTA|ifSWAP_2

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        ld              h, c            ; hl -> RR

        ld              a, [hl+]        ; a <- LOW byte of [RR]
        ld              [hl], a         ; store out LOW byte of [RR] to HIGH

        RETURN          0               ; return, enabling interrupts

        ;
        ;               Ŀ                       S Z O C
        ; SLL   RR, 1   1500
        ;                                      x x
        ;

ifSLL_1_F::                             ; Shift Logical Left 1 w/ Flags

        ; OBJ_00:       RD_GTA|ifSLL_1_F

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              a, [hl]         ; a <- LOW byte of [RR]
        add             a, a            ; shift left LOW byte of [RR] once
        ld              [hl+], a

        ld              a, [hl]         ; a <- HIGH byte of [RR]

        jr              z, .ChkZero

        adc             a, a            ; shift left HIGH byte of [RR] once
        ld              [hl], a         ; store out HIGH byte of [RR] << 1
.ChkSign
        res             2, b            ; reset Zero (bit #2)

        add             a, a            ; is sign (bit #7) set ?
        jr              nc, .NoSign

        set             3, b            ; set Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.ChkZero

        adc             a, a            ; shift left HIGH byte of [RR] once
        ld              [hl], a

        jr              nz, .ChkSign    ; if HIGH byte isn't = 0, check Sign

        set             2, b            ; set Zero (bit #2)
.NoSign
        res             3, b            ; reset Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

        ;
        ;               Ŀ                       S Z O C
        ; SLL   RR, 1   1500
        ;               
        ;

ifSLL_1::                               ; Shift Logical Left 1

        ; NOTE:  Even though SLL RR, 1 only supports R0-R3, we need to
        ;        include support for R4-R6 in the "jump entry" table
        ;        since ADDR RD, RD instructions which do not modify
        ;        flags are recoded as SLL RR, 1 instructions.

        ; OBJ_00:       RD_GTA|ifSLL_1

        RSING_TAB       l, 6, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        sla             [hl]            ; shift left LOW byte of [RR] once
        inc             l
        rl              [hl]            ; shift HIGH byte of [RR] left once

        RETURN          0               ; return, enabling interrupts

        ;
        ;               Ŀ                       S Z O C
        ; SLL   RR, 2   1500
        ;                                      x x
        ;

ifSLL_2_F::                             ; Shift Logical Left 2 w/ Flags

        ; OBJ_00:       GTA|ifSLL_2_F

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              a, [hl+]        ; a <- LOW byte of [RR]
        add             a, a            ; shift left LOW byte of [RR] once
        jr              z, .ChkZero

        ld              e, a            ; e = LOW byte of [RR]
        ld              a, [hl]         ; a <- HIGH byte of [RR]

        rla                             ; shift left HIGH byte of [RR] once
        sla             e               ; shift left LOW byte of [RR] twice
        rla                             ; shift left HIGH byte of [RR] twice

        ld              [hl-], a        ; store out HIGH byte of [RR] << 2
        ld              [hl], e         ; store out LOW byte of [RR] << 2
.ChkSign
        res             2, b            ; reset Zero (bit #2)

        add             a, a            ; is sign (bit #7) set ?
        jr              nc, .NoSign

        set             3, b            ; set Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.ChkZero

        ld              a, [hl]         ; a <- HIGH byte of [RR]
        rla                             ; shift HIGH byte of [RR] once
        add             a, a            ; shift HIGH byte of [RR] twice

        ld              [hl-], a        ; store out HIGH byte of [RR] << 2
        ld              [hl], 0         ; store out LOW byte of [RR] << 2 (0)

        jr              nz, .ChkSign

        set             2, b            ; set Zero (bit #2)
.NoSign
        res             3, b            ; reset Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

        ;
        ;               Ŀ                       S Z O C
        ; SLL   RR, 2   1500
        ;               
        ;

ifSLL_2::                               ; Shift Logical Left 2

        ; OBJ_00:       RD_GTA|ifSLL_2

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              a, [hl+]        ; a <- LOW byte of [RR]
        add             a, a            ; shift LOW byte of [RR] once
        ld              e, a            ; e = LOW byte of [RR] << 1

        ld              a, [hl]         ; a <- HIGH byte of [RR]

        rla                             ; shift HIGH byte of [RR] once
        sla             e               ; shift LOW byte of [RR] twice
        rla                             ; shift HIGH byte of [RR] twice

        ld              [hl-], a        ; store out HIGH byte of [RR] << 2
        ld              [hl], e         ; store out LOW byte of [RR] << 2

        RETURN          0               ; return, enabling interrupts

        ;              Ŀ
        ;               Ŀ  Ŀ                S Z O C
        ; RLC   RR, 1  ĴCĳ150
        ;                                   x x   x
        ;

ifRLC_1_F::                             ; Rotate Left thru Carry 1 w/ Flags

        ; OBJ_00:       RD_GTA|ifRLC_1_F

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              a, b            ; a = ..1.SZOC
        and             $23             ;   = ..1...OC (clear Sign and Zero)
        rra                             ;   = ...1...O, CY = old Carry

        rl              [hl]            ; shift LOW byte of [RD] left once
        jr              z, .ChkZero

        inc             l
        rl              [hl]            ; shift HIGH byte of [RD] left once
.ChkSign
        rla                             ; a = ..1...Oc
        ld              b, a            ; update Status

        bit             7, [hl]         ; is sign (bit #7) set ?
        ret             z               ; *

        set             3, b            ; set Sign

        RETURN          0               ; return, enabling interrupts

.ChkZero
        inc             l
        rl              [hl]            ; shift HIGH byte of [RD] left once
        jr              nz, .ChkSign

        rla                             ; a = ..1...Oc
        or              $04             ; set Zero (bit #2)
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

        ;
        ;               Ŀ  Ŀ                  S Z O C
        ; RLC   RR, 1   150ĴC
        ;                 
        ;

ifRLC_1::                               ; Rotate Left thru Carry 1

        ; OBJ_00:       RD_GTA|ifRLC_1

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              a, b            ; a = ..1.SZOC (Status)
        rra                             ; CY = Carry

        rl              [hl]            ; bit #0 = Carry, CY = bit #7
        inc             l
        rl              [hl]            ; bit #0 = old CY, new CY = bit #7

        RETURN          0               ; return, enabling interrupts

        ;              Ŀ
        ;               Ŀ  Ŀ  Ŀ           S Z O C
        ; RLC   RR, 2  ĴCĴOĳ150
        ;                                x x x x
        ;

ifRLC_2_F::                             ; Rotate Left thru Carry 2 w/ Flags

        ; OBJ_00:       RD_GTA|ifRLC_2_F

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              a, b            ; a = ..1.SZOC (Status)
        and             $23             ;   = ..1...OC (clear Sign and Zero)
        rra                             ;   = ...1...O (CY = old Carry)
        ld              e, a

        ld              a, [hl+]        ; a <- LOW byte of [RR]
        rla                             ; rot. LOW byte of [RR], bit #0 = CY

        rl              [hl]            ; rot. HIGH byte of [RR] once
        dec             l

        rr              e               ; e = c...1... (CY = old Overflow)

        adc             a, a            ; rot. LOW byte of [RR], bit #0 = CY
        ld              [hl+], a        ; store out LOW byte of [RR] << 2

        ld              a, e            ; a = c...1...

        jr              z, .ChkZero

        rl              [hl]            ; rot. HIGH byte of [RR] twice
.ChkSign
        rla                             ; a = ...1...o (CY = new Carry)
        rla                             ;   = ..1...oc
        ld              b, a            ; update Status

        bit             7, [hl]         ; is sign (bit #7) set ?
        ret             z               ; *

        set             3, b            ; set Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.ChkZero

        rl              [hl]            ; rot. HIGH byte of [RR] twice
        jr              nz, .ChkSign

        inc             a               ; set Zero
        rla                             ; a = ...1..zo (CY = new Carry)
        rla                             ;   = ..1..zoc
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

        ;
        ;               Ŀ  Ŀ  Ŀ             S Z O C
        ; RLC   RR, 2   150ĴCĴO
        ;                   
        ;

ifRLC_2::                               ; Rotate Left thru Carry 2

        ; OBJ_00:       RD_GTA|ifRLC_2

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              a, b            ; a = ..1.SZOC (Status)
        rra                             ;   = ...1.SZO (CY = old Carry)
        ld              e, a

        ld              a, [hl+]        ; a <- LOW byte of [RD]
        rla                             ; rot. LOW byte of [RD], bit #0 = CY

        rl              [hl]            ; rot. HIGH byte of [RD] once
        dec             l

        rr              e               ; e = ....1.SZ (CY = old Overflow)

        rla                             ; shift LOW byte of [RD], bit #0 = CY
        ld              [hl+], a        ; store out LOW byte of [RR] << 2

        rl              [hl]            ; rot. HIGH byte of [RD] twice

        RETURN          0               ; return, enabling interrupts

        ;
        ;               Ŀ  Ŀ                  S Z O C
        ; SLLC  RR, 1   Cĳ1500
        ;                                   x x   x
        ;

ifSLLC_1_F::                            ; Shift Logical Left thru Carry 1 w/ Flags

        ; OBJ_00:       RD_GTA|ifSLLC_1_F

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              a, b            ; a = ..1.SZOC (Status)
        and             $22             ;   = ..1...O. (isolate Overflow)

        sla             [hl]            ; shift LOW byte of [RR] once
        jr              z, .ChkZero

        inc             l
        rl              [hl]            ; rot. HIGH byte of [RR] once

.ChkSign
        adc             0               ; a = ..1...Oc
        ld              b, a            ; update Status

        bit             7, [hl]         ; is sign (bit #7) set ?
        ret             z               ; *

        set             3, b            ; set Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.ChkZero

        inc             l
        rl              [hl]            ; rot. HIGH byte of [RR] once

        jr              nz, .ChkSign

        adc             $04             ; a = ..1..zOc
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

        ;
        ;               Ŀ                       S Z O C
        ; SLLC  RR, 1   1500
        ;               
        ;

ifSLLC_1::                              ; Shift Logical Left thru Carry 1

        ; OBJ_00:       RD_GTA|ifSLL_1

        ;
        ;               Ŀ  Ŀ  Ŀ             S Z O C
        ; SLLC  RR, 2   CĳOĳ1500
        ;                                x x x x
        ;

ifSLLC_2_F::                            ; Shift Logical Left thru Carry 2 w/ Flags

        ; OBJ_00:       RD_GTA|ifSLLC_2_F

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              d, $10          ; d = ...1.SZO

        ld              a, [hl+]        ; a <- LOW byte of [RR]
        add             a, a            ; shift LOW byte of [RR] once
        jr              z, .ChkZero

        rl              [hl]            ; rot. HIGH byte of [RR] once
        dec             l

        rr              d               ; d = c...1.SZ

        add             a, a            ; shift LOW byte of [RR] twice
        ld              [hl+], a        ; store out LOW byte of [RR] << 2

        rl              [hl]            ; rot. HIGH byte of [RR] twice

        ld              a, d            ; a = c...1.SZ
        rla                             ;   = ...1.SZo (CY = new Carry)
        rla                             ;   = ..1.SZoc
        ld              b, a            ; update Status

        bit             7, [hl]         ; is sign (bit #7) set ?
        ret             z               ; *

        set             3, b            ; set Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.ChkZero

        ld              a, [hl]         ; a <- HIGH byte of [RR]
        rla                             ; shift HIGH byte of [RR] once
        rr              d               ; d = c...1.SZ
        add             a, a            ; shift HIGH byte of [RR] twice

        ld              [hl-], a        ; store out HIGH byte of [RR] << 2

        jr              z, .Zero

        ld              [hl], 0         ; store out LOW byte of [RR] << 2 (0)

        rl              d               ; d = ...1.SZo, CY = new Carry
        rl              d               ;   = ..1.SZoc
        ld              b, d            ; update Status

        add             a, a            ; is sign (bit #7) set ?
        ret             nc              ; *

        set             3, b            ; set Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.Zero

        ld              a, d            ; a = c...1.SZ
        inc             a               ; set Zero
        rla                             ;   = ...1.SZo, CY = new Carry
        rla                             ;   = ..1.SZoc
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

        ;
        ;               Ŀ                       S Z O C
        ; SLLC  RR, 2   1500
        ;               
        ;

ifSLLC_2::                              ; Shift Logical Left thru Carry 2

        ; OBJ_00:       RD_GTA|ifSLL_2

        ;
        ;                   Ŀ                   S Z O C
        ; SLR   RR, 1   0150
        ;                                      x x
        ;

ifSLR_1_F::                             ; Shift Logical Right 1 w/ Flags

        ; OBJ_00:       RD_GTA|ifSLR_1_F

        RSING_TAB       l, 3, 0, rtRD_P1  ; l <- LOW byte of RR + 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR + 1

        srl             [hl]            ; shift HIGH byte of [RR] once
        jr              c, .Sign

        res             3, b            ; reset Sign (bit #3)

        jr              nz, .NoZero

        dec             l
        rr              [hl]            ; rot. LOW byte of [RR] once

        jr              z, .Zero

        res             2, b            ; reset Zero (bit #2)

        RETURN          0               ; return, enabling interrupt

.Sign
        set             3, b            ; set Sign (bit #3)
.NoZero
        dec             l
        rr              [hl]            ; rot. LOW byte of [RR] once

        res             2, b            ; reset Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

.Zero
        set             2, b            ; set Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

        ;
        ;                   Ŀ                   S Z O C
        ; SLR   RR, 1   0150
        ;                   
        ;

ifSLR_1::                               ; Shift Logical Right 1

        ; OBJ_00:       RD_GTA|ifSLR_1

        RSING_TAB       l, 3, 0, rtRD_P1  ; l <- LOW byte of RR + 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR + 1

        srl             [hl]            ; shift HIGH byte of [RR] once
        dec             l
        rr              [hl]            ; shift LOW byte of [RR] once

        RETURN          0               ; return, enabling interrupts

        ;
        ;                   Ŀ                   S Z O C
        ; SLR   RR, 2   0150
        ;                                      x x
        ;

ifSLR_2_F::                             ; Shift Logical Right 2 w/ Flags

        ; OBJ_00:       RD_GTA|ifSLR_2_F

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              a, [hl+]        ; a <- LOW byte of [RR]
        ld              d, [hl]         ; d <- HIGH byte of [RR]

        srl             d               ; shift HIGH byte of [RR] once
        jr              z, .ChkZero

        rra                             ; rot. LOW byte of [RR] once
        srl             d               ; shift HIGH byte of [RR] twice
        jr              c, .Sign

        rra                             ; rot. LOW byte of [RR] twice
.NoSign
        ld              [hl], d         ; store out HIGH byte of [RR] >> 2
        dec             l
        ld              [hl], a         ; store out LOW byte of [RR] >> 2

        res             2, b            ; reset Zero (bit #2)
        res             3, b            ; reset Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.Sign

        rra                             ; rot. LOW byte of [RR] twice

        ld              [hl], d         ; store out HIGH byte of [RR] >> 2
        dec             l
        ld              [hl], a         ; store out LOW byte of [RR] >> 2

        res             2, b            ; reset Zero (bit #2)
        set             3, b            ; set Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.ChkZero

        rra                             ; rot. LOW byte of [RR] once
        srl             a               ; rot. LOW byte of [RR] twice
        jr              nz, .NoSign

        dec             l               ; pos @ LOW of [RR] (HIGH already 0)
        ld              [hl], a         ; store out LOW byte of [RR] >> 2 (0)

        set             2, b            ; set Zero (bit #2)
        res             3, b            ; reset Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

        ;
        ;                   Ŀ                   S Z O C
        ; SLR   RR, 2   0150
        ;                   
        ;

ifSLR_2::                               ; Shift Logical Right 2

        ; OBJ_00:       RD_GTA|ifSLR_2

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              a, [hl+]        ; a <- LOW byte of [RR]

        srl             [hl]            ; shift HIGH byte of [RR] once
        rra                             ; rot. LOW byte of [RR] once
        srl             [hl]            ; shift HIGH byte of [RR] twice
        rra                             ; rot. LOW byte of [RR] twice

        dec             l
        ld              [hl], a         ; store out LOW byte of [RR]

        RETURN          0               ; return, enabling interrupts

        ;
        ;                   Ŀ                   S Z O C
        ; SAR   RR, 1   150
        ;                                     x x
        ;               

ifSAR_1_F::                             ; Shift Arithmetic Right 1 w/ Flags

        ; OBJ_00:       RD_GTA|ifSAR_1_F

        RSING_TAB       l, 3, 0, rtRD_P1  ; l <- LOW byte of RR + 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR + 1

        sra             [hl]            ; shift HIGH byte of [RR] once
        jr              c, .Sign

        res             3, b            ; reset Sign (bit #2)

        jr              nz, .NoZeroHi

        dec             l
        rr              [hl]            ; rot. LOW byte of [RR] once
        jr              nz, .NoZeroLo

        set             2, b            ; set Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

.Sign
        set             3, b            ; set Sign (bit #3)
.NoZeroHi
        dec             l
        rr              [hl]            ; rot. LOW byte of [RR] once
.NoZeroLo
        res             2, b            ; reset Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

        ;
        ;                   Ŀ                   S Z O C
        ; SAR   RR, 1   150
        ;                  
        ;               

ifSAR_1::                               ; Shift Arithmetic Right 1

        ; OBJ_00:       RD_GTA|ifSAR_1

        RSING_TAB       l, 3, 0, rtRD_P1  ; l <- LOW byte of RR + 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR + 1

        sra             [hl]            ; shift HIGH once, CY = bit #0
        dec             l
        rr              [hl]            ; rot. LOW once, bit #7 = CY

        RETURN          0               ; return, enabling interrupts

        ;
        ;                   Ŀ                   S Z O C
        ; SAR   RR, 2   150
        ;                                     x x
        ;               

ifSAR_2_F::                             ; Shift Arithmetic Right 2 w/ Flags

        ; OBJ_00:       RD_GTA|ifSAR_2_F

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              a, [hl+]        ; a <- LOW byte of [RR]
        ld              d, [hl]         ; d <- HIGH byte of [RR]

        sra             d               ; shift HIGH byte of [RR] once
        jr              z, .ChkZero

        rra                             ; rot. LOW byte of [RR] once
        sra             d               ; shift HIGH byte of [RR] twice
        jr              c, .Sign

        rra                             ; rot. LOW byte of [RR] twice
.NoSign
        ld              [hl], d         ; store out HIGH byte of [RR] >> 2
        dec             l
        ld              [hl], a         ; store out LOW byte of [RR] >> 2

        res             3, b            ; reset Sign (bit #3)
        res             2, b            ; reset Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

.Sign
        rra                             ; rot. LOW byte of [RR] twice

        ld              [hl], d         ; store out HIGH byte of [RR] >> 2
        dec             l
        ld              [hl], a         ; store out LOW byte of [RR] >> 2

        set             3, b            ; set Sign (bit #3)
        res             2, b            ; reset Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

.ChkZero

        rra                             ; rot. LOW byte of [RR] once
        srl             a               ; rot. LOW byte of [RR] twice (test Z)
        jr              nz, .NoSign

        dec             l               ; pos @ LOW of [RR] (HIGH already 0)
        ld              [hl], a         ; store out LOW byte of [RR] >> 2 (0)

        res             3, b            ; reset Sign (bit #3)
        set             2, b            ; set Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

        ;
        ;                   Ŀ                   S Z O C
        ; SAR   RR, 2   150
        ;                  
        ;               

ifSAR_2::                               ; Shift Arithmetic Right 2

        ; OBJ_00:       RD_GTA|ifSAR_2

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              a, [hl+]        ; a <- LOW byte of [RR]

        sra             [hl]            ; shift HIGH 1c, CY = bit #0
        rra                             ; rot. LOW 1c, bit #7 = CY
        sra             [hl]            ; shift HIGH 2x, CY = bit #0
        rra                             ; rot. LOW 2x, bit #7 = CY

        dec             l
        ld              [hl], a         ; store LOW byte of [RR] >> 2 + bits

        RETURN          0               ; return, enabling interrupts

        ;               Ŀ
        ;                 Ŀ  Ŀ              S Z O C
        ; RRC   RR, 1   C150
        ;                                   x x   x
        ;

ifRRC_1_F::                             ; Rotate Right thru Carry w/ Flags

        ; OBJ_00:       RD_GTA|ifRRC_1_F

        RSING_TAB       l, 3, 0, rtRD_P1  ; l <- LOW byte of RR + 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR + 1

        ld              a, b            ; a = ..1.SZOC (Status)
        and             $23             ;   = ..1...OC (clear Sign and Zero)
        rra                             ;   = ...1...O (CY = old Carry)

        rr              [hl]            ; rot. HIGH byte of [RR] once
        jr              c, .Sign
        jr              z, .ChkZero

        dec             l
        rr              [hl]            ; rot. LOW byte of [RR] once
.NoSign
        rla                             ; a = ..1...Oc
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

.Sign

        dec             l
        rr              [hl]            ; rot. LOW byte of [RR] once

        rla                             ; a = ..1...Oc
        or              $08             ; set Sign (bit #3)
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

.ChkZero

        dec             l
        rr              [hl]            ; rot. LOW byte of [RR] once
        jr              nz, .NoSign

        rla                             ; a = ..1...Oc
        or              $04             ; set Zero (bit #2)
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

        ;
        ;               Ŀ  Ŀ                  S Z O C
        ; RRC   RR, 1   C150
        ;                 
        ;

ifRRC_1::                               ; Rotate Right thru Carry 1

        ; OBJ_00:       RD_GTA|ifRRC_1

        RSING_TAB       l, 3, 0, rtRD_P1  ; l <- LOW byte of RR + 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR + 1

        ld              a, b            ; a = ..1.SZOC (Status)
        rra                             ;   = ...1.SZO (CY = Carry)

        rr              [hl]            ; rot. HIGH, bit #7 = CY, CY = bit #0
        dec             l               ; position @ LOW byte of [RR]
        rr              [hl]            ; rot. LOW, bit #7 = CY, CY = bit #0

        RETURN          0               ; return, enabling interrupts

        ;               Ŀ
        ;                 Ŀ  Ŀ  Ŀ         S Z O C
        ; RRC   RR, 2   OC150
        ;                                x x x x
        ;

ifRRC_2_F::                             ; Rotate Right thru Carry 2 w/ Flags

        ; OBJ_00:       RD_GTA|ifRRC_2_F

        RSING_TAB       l, 3, 0, rtRD_P1  ; l <- LOW byte of RR + 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR + 1

        ld              a, b            ; a = ..1.SZOC (Status)
        and             $23             ;   = ..1...OC (clear Sign and Zero)
        rra                             ;   = ...1...O (CY = old Carry)

        ld              d, [hl]         ; d <- HIGH byte of [RR]
        dec             l

        rr              d               ; rot. HIGH byte of [RR] once
        rr              [hl]            ; rot. LOW byte of [RR] once
        rra                             ; a = c...1... (CY = old Overflow)
        rr              d               ; rot. HIGH byte of [RR] twice

        jr              c, .Sign
        jr              z, .ChkZero

        rr              [hl]            ; rot. LOW byte of [RR] twice
.NoSign
        inc             l
        ld              [hl], d         ; store out HIGH byte of [RR] >> 2

        rla                             ; a = ...1...o, CY = new Carry
        rla                             ;   = ..1...oc
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

.Sign
        rr              [hl]            ; rot. LOW byte of [RR] twice

        inc             l
        ld              [hl], d         ; store out HIGH byte of [RR] >> 2

        rla                             ; a = ...1...o, CY = new Carry
        rla                             ;   = ..1...oc
        or              $08             ; set Sign (bit #3)
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

.ChkZero

        rr              [hl]            ; rot. LOW byte of [RR] twice
        jr              nz, .NoSign

        inc             a               ; set Zero
        rla                             ; a = ...1..zo, CY = new Carry
        rla                             ;   = ..1..zoc
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

        ;
        ;               Ŀ  Ŀ  Ŀ             S Z O C
        ; RRC   RR, 2   OC150
        ;                   
        ;

ifRRC_2::                               ; Rotate Right thru Carry 2

        ; OBJ_00:       RD_GTA|ifRRC_2

        RSING_TAB       l, 3, 0, rtRD_P1  ; l <- LOW byte of RR + 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR + 1

        ld              a, b            ; a = ..1.SZOC (Status)
        rra                             ;   = ...1.SZO (CY = old Carry)
        ld              d, a

        ld              a, [hl-]        ; a <- HIGH byte of [RR]

        rra
        rr              [hl]
        rr              d               ; d = c...1.SZ (CY = old Overflow)
        rra
        rr              [hl]

        inc             l
        ld              [hl], a         ; store out HIGH byte of [RR] >> 2

        RETURN          0               ; return, enabling interrupts

        ;
        ;                   Ŀ  Ŀ              S Z O C
        ; SARC  RR, 1   15             0C
        ;                                  x x   x
        ;               

ifSARC_1_F::                            ; Shift Arithmetic Right thru Carry 1 w/ Flags

        ; OBJ_00:       RD_GTA|ifSARC_1_F

        RSING_TAB       l, 3, 0, rtRD_P1  ; l <- LOW byte of RR + 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR + 1

        ld              a, b            ; a = ..1.SZOC (Status)
        and             $23             ;   = ..1...OC (clear Sign and Zero)
        rra                             ;   = ...1...O

        sra             [hl]            ; shift HIGH byte of [RR] once
        jr              c, .Sign
        jr              z, .ChkZero

        dec             l
        rr              [hl]            ; rot. LOW byte of [RR] once
.NoSign
        rla                             ; a = ..1...Oc
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

.Sign
        dec             l
        rr              [hl]            ; rot. LOW byte of [RR] once

        rla                             ; a = ..1...Oc
        or              $08             ; set Sign (bit #3)
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

.ChkZero

        dec             l
        rr              [hl]            ; rot. LOW byte of [RR] once
        jr              nz, .NoSign

        rla                             ; a = ..1...Oc
        or              $04             ; set Zero (bit #2)
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

        ;
        ;                   Ŀ                   S Z O C
        ; SARC  RR, 1   15             0
        ;                  
        ;               

ifSARC_1::                              ; Shift Arithmetic Right thru Carry 1

        ; OBJ_00:       RD_GTA|ifSAR_1

        ;
        ;                   Ŀ  Ŀ  Ŀ         S Z O C
        ; SARC  RR, 2   15             0OC
        ;                               x x x x
        ;               

ifSARC_2_F::                            ; Shift Arithmetic Right thru Carry 2 w/ Flags

        ; OBJ_00:       RD_GTA|ifSARC_2_F

        RSING_TAB       l, 3, 0, rtNone ; l <- LOW byte of RR

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RR

        ld              d, $10          ; d = ...1.SZO

        ld              a, [hl+]        ; a <- LOW byte of [RR]

        sra             [hl]            ; shift HIGH byte of [RR] once
        rra                             ; shift LOW byte of [RR] once

        rr              d               ; d = c...1.SZ

        sra             [hl]            ; shift HIGH byte of [RR] twice
        jr              c, .Sign
        jr              z, .ChkZero

        rra                             ; shift LOW byte of [RR] twice
.NoSign
        dec             l
        ld              [hl], a         ; store out LOW byte [RR] >> 2

        ld              a, d
        rla                             ; a = ...1...o, CY = new Carry
        rla                             ;   = ..1...oc
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

.Sign
        rra                             ; shift LOW byte of [RR] twice

        dec             l
        ld              [hl], a         ; store out LOW byte [RR] >> 2

        ld              a, d
        rla                             ; a = ...1...o, CY = new Carry
        rla                             ;   = ..1...oc
        or              $08             ; set Sign (bit #3)
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

.ChkZero

        rr              a               ; shift LOW byte of [RR] 2x (test Z)
        jr              nz, .NoSign

        dec             l
        ld              [hl], a         ; store out LOW byte [RR] >> 2

        ld              a, d
        inc             a               ; set Zero (bit #2)
        rla                             ; a = ...1..zo, CY = new Carry
        rla                             ;   = ..1..zoc
        ld              b, a            ; update Status

        RETURN          0               ; return, enabling interrupts

        ;
        ;                   Ŀ                   S Z O C
        ; SARC  RR, 2   15             0
        ;                  
        ;               

ifSARC_2::                              ; Shift Arithmetic Right thru Carry 2

        ; OBJ_00:       RD_GTA|ifSAR_2

; ---------------------------------------------------------------------------
;     .::** CP1610's register-register move & operate instructions **::.
; ---------------------------------------------------------------------------

ifMOVR_F::                              ; MOVe Register w/ Flags

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RSRD_GTA|ifMOVR_F.Tail

        RPAIR_HEAD      hl, 5, 0, 5, 0, rtOmitSame | rtTail

        ld              e, h            ; save LOW byte of RD in e
        ld              h, c            ; hl -> RS

        ld              a, [hl+]        ; a <- LOW byte of [RS]
        ld              d, [hl]         ; d <- HIGH byte of [RS]

        ld              l, e            ; hl -> RD

        ld              [hl+], a        ; store out LOW byte of [RS] to [RD]
        ld              [hl], d         ; store out HIGH byte of [RS] to [RD]

        or              d               ; are both LOW & HIGH bytes = 0 ?
        jr              z, .Zero

        res             2, b            ; reset Zero (bit #2)

        add             a, a            ; is sign (bit #7) set ?
        jr              c, .ChkSign
.NoSign
        res             3, b            ; reset Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.Zero

        set             2, b            ; set Zero (bit #2)
        res             3, b            ; reset Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.ChkSign

        bit             7, [hl]         ; is sign (bit #15) set ?
        jr              z, .NoSign

        set             3, b            ; set Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

        RPAIR_TAIL      hl, rtOmitSame

ifMOVR::                                ; MOVe Register

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RSRD_GTA|ifMOVR.Tail

        RPAIR_HEAD      hl, 6, 0, 6, 0, rtOmitSame | rtOmitLast | rtTail

        DOREG           0               ; da <- [RS], hl -> RD

        ld              [hl+], a        ; store out LOW byte of [RS] to [RD]
        ld              [hl], d         ; store out HIGH byte of [RS] to [RD]

        RETURN          0               ; return, enabling interrupts

        RPAIR_TAIL      hl, rtOmitSame

ifMOVR_R0_R1::                          ; MOVe Register [R0] to [R1]

        ; NOTE:  This is a special, optimized version of the instruction
        ;        MOVR R0, R1 since it occurs VERY frequently.  It only
        ;        handles the one combination of registers.
        ;
        ;        This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       ifMOVR_R0_R1

        ld              hl, R0          ; hl -> R0

        ld              a, [hl+]        ; a <- LOW byte of [R0]
        ld              d, [hl]         ; d <- HIGH byte of [R0]

        inc             l               ; hl -> R1

        ld              [hl+], a        ; store out LOW byte of [R0]
        ld              [hl], d         ; store out HIGH byte of [R0]

        RETURN          0               ; return, enabling interrupts

        ; ^ 16 cycles, 9 bytes

ifJR::                                  ; Jump Register

        ; NOTE:  It's assumed that the flags (Status) would never be
        ;        tested as a result of a JR, thus there exists only
        ;        one version of this instruction.

        ; OBJ_00:       RS_GTA|ifJR

        RSING_TAB       l, 0, 5, rtRS_P1  ; l <- LOW byte of RD + 1

ifJR_1::

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RS + 1

        ld              a, [hl-]        ; a = HIGH byte of [RS]
        ld              e, [hl]         ; e = LOW byte of [RS]

        JUMP            mpR0inH | mpJAinAE

;ifJR_R6::
;
;        ; OBJ_00:       ifJR_R6
;
;        ld              l, R6 & $FF + 1  ; l <- LOW byte of R6 + 1
;
;        jr              ifJR_1           ; uncommon, so let's share code

ifJR_R7::

        ; OBJ_00:       ifNOP

ifJR_NBS::                              ; Jump Register w/ No Bank Switching

        ; NOTE:  It's assumed that the flags (Status) would never be
        ;        tested as a result of a JR, thus there exists only
        ;        one version of this instruction.

        ; OBJ_00:       RS_GTA|ifJR_NBS

        RSING_TAB       l, 0, 5, rtNone ; l <- LOW byte of RS

ifJR_NBS_1::                            ; entry point for ifJR_R6_NBS

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RS

        ld              a, [hl+]        ; a <- LOW byte of [RS]
        ld              l, [hl]         ; l <- HIGH byte of [RS]

        inc             h
        ld              h, [hl]         ; h <- HIGH of JumpAddr (Real) >> 1

        ld              l, a
        add             hl, hl

        ld              sp, hl          ; sp = JumpAddr (Real)

        RETURN          0               ; return, enabling interrupts

        ; ^ 2/5 + 14 + 4 = 20/23 cycles vs. 32/35 cycles (saves 12 cycles)

;ifJR_R6_NBS::
;
;        ; OBJ_00:       ifJR_R6_NBS
;
;        ld              l, R6 & $FF     ; l <- LOW byte of R6
;
;        jr              ifJR_NBS_1

ifMOVR_SR7::                            ; MOVe Register (Source) [R7]

        ; NOTE:  It's assumed that the flags (Status) would never be
        ;        tested as a result of a MOVR R7, RD, thus there exists
        ;        only one version of this instruction.
        ;
        ;        This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RD_GTA|ifMOVR_SR7

        RSING_TAB       e, 0, 5, rtNone ; e <- LOW byte of RD

ifMOVR_SR7_1::

        ld              d, c            ; de -> RD

        ldhl            sp, 0           ; hl = [R7] (Real)

        srl             h
        ld              a, l
        rra                             ; ha = [R7] (Real) >> 1

        ld              [de], a         ; store out LOW byte of [R7] to [RD]
        inc             e

        ldh             a, [R7_Conv]
        add             h
        ld              [de], a         ; store out HIGH byte of [R7] to [RD]

        RETURN          0               ; return, enabling interrupts

ifTSTR::                                ; TeST Register

        ; OBJ_00:       ifNOP

ifTSTR_F::                              ; TeST Register w/ Flags

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RD_GTA|ifTSTR_F

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RS

        ld              h, c            ; hl -> RS

        ld              a, [hl+]        ; a <- LOW byte of [RS]

        or              [hl]            ; is [RS] = 0 ?
        jr              z, .Zero

        res             2, b            ; reset Zero (bit #2)

        bit             7, [hl]         ; is sign (bit #7) set ?
        jr              nz, .Sign

        res             3, b            ; reset Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.Sign

        set             3, b            ; set Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

.Zero

        set             2, b            ; set Zero (bit #2)
        res             3, b            ; reset Sign (bit #3)

        RETURN          0               ; return, enabling interrupts

; ---------------------------------------------------------------------------
;          .::** Code for ADDR RS, RD (with and without flags) **::.
; ---------------------------------------------------------------------------

ifADDR_F::                              ; ADD Register w/ Flags

        ; OBJ_00:       RSRD_GTA|ifADDR_F.Tail

        RPAIR_HEAD      hl, 5, 0, 5, 0, rtTail  ; l <- LOW byte of RS
                                                ; h <- LOW byte of RD

        DOREG           mpDI            ; da <- [RS], hl -> RD
        ADD_RD_F        0               ; [RD] <- [RD] + da, set Status

        RPAIR_TAIL      hl, rtNone      ; l <- LOW byte of RS
                                        ; h <- LOW byte of RD

ifADDR::                                ; ADD Register

        ; OBJ_00:       RSRD_GTA|ifADDR.Tail

        RPAIR_HEAD      hl, 5, 0, 6, 0, rtOmitSame | rtTail  ; l <- LOW of RS
                                                             ; h <- LOW of RD

        DOREG           mpDI            ; da <- [RS], hl -> RD
        ADD_RD          0               ; [RD] <- [RD] + da

        RPAIR_TAIL      hl, rtOmitSame  ; l <- LOW byte of RS
                                        ; h <- LOW byte of RD

ifADDR_DR7::                            ; ADD Register (Destination) [R7]

        ; OBJ_00:       RS_GTA|ifADDR_DR7

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RS

        DOREG_DR7                       ; da <- [RS], hl = [R7] (Real)

        srl             h
        rr              l               ; hl = [R7] (Real) >> 1

        add             l               ; a = LOW byte of [R7] + [RS]
        ld              e, a            ; save it in e (for Jump_Op)

        ldh             a, [R7_Conv]
        adc             h
        add             d               ; a = HIGH byte of [R7] + [RS]

Jump_Op:

        JUMP            mpJAinAE        ; conv. addr, switch banks if needed

ifADDR_SR7::                            ; ADD Register (Source) [R7]

        ; OBJ_00:       RD_GTA|ifADDR_SR7

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RS

        DOREG_SR7                       ; hl -> RD, da = [R7] (real) >> 1

        add             [hl]            ; a = LOW byte of [RD] + [R7]
        ld              [hl+], a        ; store out LOW byte of [RD] + [R7]

        ldh             a, [R7_Conv]
        adc             d
        add             [hl]            ; a = HIGH byte of [RD] + [R7]
        ld              [hl], a         ; store out HIGH byte of [RD] + [R7]

        RETURN          0               ; return, enabling interrupts

; ---------------------------------------------------------------------------
;        .::** Code for SUBR RS, RD (with and without flags) **::.
; ---------------------------------------------------------------------------

ifSUBR_F::                              ; SUBtract Register w/ Flags

        ; OBJ_00:       RSRD_GTA|ifSUBR_F.Tail

        RPAIR_HEAD      hl, 5, 0, 5, 0, rtOmitSame | rtTail  ; l <- LOW of RS
                                                             ; h <- LOW of RD

        DOREG           mpDI            ; da <- [RS], hl -> RD
        ld              e, a            ; e = LOW byte of [RS] (for SUB_RD_F)
        SUB_RD_F        mpSub           ; [RD] <- [RD] - de, set Status

        RPAIR_TAIL      hl, rtOmitSame  ; l <- LOW byte of RS
                                        ; h <- LOW byte of RD

ifSUBR_R6_R6_F::                        ; SUBtract [R6] from [R6] w/ Flags

        ; OBJ_00:       RD_GTA|ifSUBR_SAME_F

ifSUBR_R5_R5_F::                        ; SUBtract [R5] from [R5] w/ Flags

        ; OBJ_00:       RD_GTA|ifSUBR_SAME_F

ifSUBR_R4_R4_F::                        ; SUBtract [R4] from [R4] w/ Flags

        ; OBJ_00:       RD_GTA|ifSUBR_SAME_F

ifSUBR_R3_R3_F::                        ; SUBtract [R3] from [R3] w/ Flags

        ; OBJ_00:       RD_GTA|ifSUBR_SAME_F

ifSUBR_R2_R2_F::                        ; SUBtract [R2] from [R2] w/ Flags

        ; OBJ_00:       RD_GTA|ifSUBR_SAME_F

ifSUBR_R1_R1_F::                        ; SUBtract [R1] from [R1] w/ Flags

        ; OBJ_00:       RD_GTA|ifSUBR_SAME_F

ifSUBR_R0_R0_F::                        ; SUBtract [R0] from [R0] w/ Flags

        ; OBJ_00:       RD_GTA|ifSUBR_SAME_F

ifSUBR_SAME_F::                         ; SUBtract Register (Same) w/ Flags

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        ld              h, c            ; hl -> RD

        xor             a
        ld              [hl+], a        ; zero LOW byte of [RD]
        ld              [hl], a         ; zero HIGH byte of [RD]

ifCMPR_SAME_F::                         ; CoMPare Register (Same) w/ Flags

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ld              b, $25          ; SZOC = .1.1

        RETURN          0               ; return, enabling interrupts

        ; ^ 33 bytes (vs. 35 bytes needed to include in SUBR_F, above)

ifSUBR::                                ; SUBtract Register

        ; OBJ_00:       RSRD_GTA|ifSUBR.Tail

        RPAIR_HEAD      hl, 5, 0, 5, 0, rtOmitSame | rtTail  ; l <- LOW of RS
                                                             ; h <- LOW of RD

        DOREG           mpDI            ; da <- [RS], hl -> RD
        ld              e, a            ; e = LOW byte of [RS] (for SUB_RD)
        SUB_RD          0               ; [RD] <- [RD] - de

        RPAIR_TAIL      hl, rtOmitSame  ; l <- LOW byte of RS
                                        ; h <- LOW byte of RD

ifSUBR_DR7::                            ; SUBtract Reg. (Destination) [R7]

        ; OBJ_00:       RS_GTA|ifSUBR_DR7

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RS

        DOREG_DR7                       ; da <- [RS], hl = [R7] (real)

        ld              e, a            ; e = LOW byte of [RS]

        srl             h
        ld              a, l
        rra                             ; ha = [R7] (real) >> 1

        sub             e               ; a = LOW byte of [R7] - [RS]
        ld              e, a            ; save it in e (for Jump_Op)

        ldh             a, [R7_Conv]
        sbc             d               ;
        add             h               ; a = HIGH byte of [R7] - [RS]

        jp              Jump_Op         ; convert addr, switch bank if needed

ifSUBR_SR7::                            ; SUBtract Register (Source) [R7]

        ; OBJ_00:       RD_GTA|ifSUBR_SR7

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DOREG_SR7                       ; hl -> RD, da <- [R7] (real) >> 1

        ld              e, a

        ldh             a, [R7_Conv]
        add             d
        ld              d, a            ; de = [R7] (CP1610)

        ld              a, [hl]         ; a <- LOW byte of [RD]
        sub             e
        ld              [hl+], a        ; store out LOW byte of [RD] - [R7]

        ld              a, [hl]         ; a <- HIGH byte of [RD]
        sbc             d
        ld              [hl], a         ; store out HIGH byte of [RD] - [R7]

        RETURN          0               ; return to .Px_Loop or .Interrupt

; ---------------------------------------------------------------------------
;            .::** Code for CMPR RS, RD (with flags only) **::.
; ---------------------------------------------------------------------------

ifCMPR::                                ; CoMPare Register

        ; OBJ_00:       ifNOP

ifCMPR_F::                              ; CoMPare Register w/ Flags

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RSRD_GTA|ifCMPR_F.Tail

        RPAIR_HEAD      hl, 6, 0, 6, 0, rtOmitSame | rtTail  ; l <- LOW of RS
                                                             ; h <- LOW of RD

        DOREG           0               ; da <- [RS], hl -> RD
        ld              e, a            ; e = LOW byte of [RS]
        SUB_RD_F        mpCmp           ; [RD] - de, set Status

        RPAIR_TAIL      hl, rtOmitSame  ; l <- LOW byte of RS
                                        ; h <- LOW byte of RD

ifCMPR_R6_R7::                          ; CoMPare [R6] to [R7] w/ Flags *

        ; NOTE:  Technically, this should be labeled ifCMPR_R6_R7_F,
        ;        but BIN2IGB always expects instructions involving
        ;        R7 not to affect the flags.

        ; OBJ_00:       ifCMPR_DIFF_F

ifCMPR_R6_R6_F::                        ; CoMPare [R6] to [R6] w/ Flags

        ; OBJ_00:       ifCMPR_SAME_F

ifCMPR_R5_R5_F::                        ; CoMPare [R5] to [R5] w/ Flags

        ; OBJ_00:       ifCMPR_SAME_F

ifCMPR_R4_R4_F::                        ; CoMPare [R4] to [R4] w/ Flags

        ; OBJ_00:       ifCMPR_SAME_F

ifCMPR_R3_R3_F::                        ; CoMPare [R3] to [R3] w/ Flags

        ; OBJ_00:       ifCMPR_SAME_F

ifCMPR_R2_R2_F::                        ; CoMPare [R2] to [R2] w/ Flags

        ; OBJ_00:       ifCMPR_SAME_F

ifCMPR_R1_R1_F::                        ; CoMPare [R1] to [R1] w/ Flags

        ; OBJ_00:       ifCMPR_SAME_F

ifCMPR_R0_R0_F::                        ; CoMPare [R0] to [R0] w/ Flags

        ; OBJ_00:       ifCMPR_SAME_F

; ---------------------------------------------------------------------------
;        .::** Code for ANDR RS, RD (with and without flags) **::.
; ---------------------------------------------------------------------------

ifANDR_F::                              ; AND Register w/ Flags

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RSRD_GTA|ifANDR_F.Tail

        RPAIR_HEAD      hl, 5, 0, 5, 0, rtOmitSame | rtTail  ; l <- LOW of RS
                                                             ; h <- LOW of RD

        DOREG           0               ; da <- [RS], hl -> RD
        AND_RD_F        0               ; [RD] <- [RD] & da, set Status

        RPAIR_TAIL      hl, rtOmitSame  ; l <- LOW byte of RS
                                        ; h <- LOW byte of RD

ifANDR::                                ; AND Register w/ Flags

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RSRD_GTA|ifANDR.Tail

        RPAIR_HEAD      hl, 5, 0, 5, 0, rtOmitSame | rtTail  ; l <- LOW of RS
                                                             ; h <- LOW of RD

        DOREG           0               ; da <- [RS], hl -> RD
        AND_RD          0               ; [RD] <- [RD] & da

        RPAIR_TAIL      hl, rtOmitSame  ; l <- LOW byte of RS
                                        ; h <- LOW byte of RD

; ---------------------------------------------------------------------------
;        .::** Code for XORR RS, RD (with and without flags) **::.
; ---------------------------------------------------------------------------

ifXORR_F::                              ; XOR Register w/ Flags

        ; OBJ_00:       RSRD_GTA|ifXORR_F.Tail

        RPAIR_HEAD      hl, 5, 0, 5, 0, rtOmitSame | rtTail  ; l <- LOW of RS
                                                             ; h <- LOW of RD

        DOREG           mpDI            ; da <- [RS], hl -> RD
        XOR_RD_F        0               ; [RD] <- [RD] ^ da, set Status

        RPAIR_TAIL      hl, rtOmitSame  ; l <- LOW byte of RS
                                        ; h <- LOW byte of RD

ifXORR::                                ; XOR Register

        ; OBJ_00:       RSRD_GTA|ifXORR.Tail

        RPAIR_HEAD      hl, 5, 0, 5, 0, rtOmitSame | rtTail  ; l <- LOW of RS
                                                             ; h <- LOW of RD

        DOREG           mpDI            ; da <- [RS], hl -> RD
        XOR_RD          0               ; [RD] <- [RD] ^ da

        RPAIR_TAIL      hl, rtOmitSame  ; l <- LOW byte of RS
                                        ; h <- LOW byte of RD

ifCLRR::                                ; CLeaR Register

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RD_GTA|ifCLRR

        RSING_TAB       l, 6, 0, rtNone ; l <- LOW byte of RD

        ld              h, c            ; hl -> RD

        xor             a
        ld              [hl+], a        ; clear LOW byte of [RD]
        ld              [hl], a         ; clear HIGH byte of [RD]

        RETURN          0               ; return, enabling interrupts

ifCLRR_F::                              ; CLeaR Register w/ Flags

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.
        ;
        ;        This version of the instruction (that sets flags,
        ;        by clearing Sign and setting Zero), IS actually
        ;        needed.  See FROGGER @ $5CE4.

        ; OBJ_00:       RD_GTA|ifCLRR_F

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        ld              h, c            ; hl -> RD

        xor             a
        ld              [hl+], a        ; clear LOW byte of [RD]
        ld              [hl], a         ; clear HIGH byte of [RD]

        res             3, b            ; reset Sign (bit #3)
        set             2, b            ; set Zero (bit #2)

        RETURN          0               ; return, enabling interrupts

; ---------------------------------------------------------------------------
;                        .::** Code for ifBcond **::.
; ---------------------------------------------------------------------------

ifB::                                   ; Unconditional

        ; OBJ_00:       ifB
        ; OBJ_01:       XX .SHL. 8 + YY

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl
        ld              sp, hl

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     C      C = 1                          0001     
        ;    LGT     Carry                                   
        ;            (Logical Greater Than)                  
        ; 

ifBC::                                  ; Carry (Logical Greater Than)

        ; OBJ_00:       ifBC
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        rrca                            ; is C = 1 ?
        ret             nc              ; *

        ld              sp, hl          ; sp = OBJ_01 (Jump Address)

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     OV     O = 1                          0010     
        ;            Overflow                                
        ; 

ifBOV::                                 ; Overflow

        ; OBJ_00:       ifBOV
        ; OBJ_01:       XX .SHL. 8 + YY

        bit             1, b            ; is O = 1 ?

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        ret             z               ; *

        ld              sp, hl          ; sp = OBJ_01 (Jump Address)

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     PL     S = 0                          0011     
        ;            Plus                                    
        ; 

ifBPL::                                 ; Plus

        ; OBJ_00:       ifBPL
        ; OBJ_01:       XX .SHL. 8 + YY

        bit             3, b            ; is S = 0 ?

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        ret             nz              ; *

        ld              sp, hl          ; sp = OBJ_01 (Jump Address)

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     ZE     Z = 1                          0100     
        ;     EQ     Zero (Equal)                            
        ; 

ifBZE::                                 ; Zero (Equal)

        ; OBJ_00:       ifBZE
        ; OBJ_01:       XX .SHL. 8 + YY

        bit             2, b            ; is Z = 1 ?

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        ret             z               ; *

        ld              sp, hl          ; sp = OBJ_01 (Jump Address)

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     LT     S xor O = 1                    0101     
        ;            Less Than                               
        ; 

ifBLT::                                 ; Less Than

        ; OBJ_00:       ifBLT
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        add             a, a            ; a = .1.SZOC.
        add             a, a            ;   = 1.SZOC.. (put O in bit #3)
        xor             b

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        and             $08             ; is S xor O = 1 ?
        ret             z               ; *

        ld              sp, hl          ; sp = OBJ_01 (Jump Address)

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     LE     Z or ( S xor O ) = 1           0110     
        ;            Less than or Equal                      
        ; 

ifBLE::                                 ; Less than or Equal

        ; OBJ_00:       ifBLE
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        add             a, a            ; a = .1.SZOC.
        add             a, a            ;   = 1.SZOC.. (put O in bit #3)
        xor             b               ; put S xor O in bit #3
        rrca                            ;             in bit #2
        or              b               ; put Z or ( S xor O ) bit #2

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        and             $04             ; is Z or ( S xor O ) = 1 ?
        ret             z               ; *

        ld              sp, hl          ; sp = OBJ_01 (Jump Address)

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;    USC     C xor S = 1                    0111     
        ;            Unequal Sign and Carry                  
        ; 

ifBUSC::                                ; Unequal Sign and Carry

        ; OBJ_00:       ifBUSC
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        swap            a               ;   = SZOC..1.
        rlca                            ;   = ZOC..1.S (put S in bit #0)
        xor             b               ; put C xor S in bit #0

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        rrca                            ; is C xor S = 1 ? (put in CY)
        ret             nc

        ld              sp, hl          ; sp = OBJ_01 (Jump Address)

        RETURN          0               ; return, enabling interrupts

ifNOPP::                                ; No OPeration

        ; OBJ_00:       ifB
        ; OBJ_01:       XX .SHL. 8 + YY

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     NC     C = 0                          1001     
        ;    LLT     No Carry                                
        ;            (Logical Less Than)                     
        ; 

ifBNC::                                 ; No Carry (Logical Less Than)

        ; OBJ_00:       ifBNC
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        rrca                            ; is C = 0 ? (put in CY)
        ret             c

        ld              sp, hl          ; sp = OBJ_01 (Jump Address)

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;    NOV     O = 0                          1010     
        ;            No Overflow                             
        ; 

ifBNOV::                                ; No Overflow

        ; OBJ_00:       ifBNOV
        ; OBJ_01:       XX .SHL. 8 + YY

        bit             1, b            ; is O = 0 ?

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        ret             nz

        ld              sp, hl          ; sp = OBJ_01 (Jump Address)

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     MI     S = 1                          1011     
        ;            Minus                                   
        ; 

ifBMI::                                 ; Minus

        ; OBJ_00:       ifBMI
        ; OBJ_01:       XX .SHL. 8 + YY

        bit             3, b            ; is S = 1 ?

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        ret             z

        ld              sp, hl          ; sp = OBJ_01 (Jump Address)

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;    NZE     Z = 0                          1100     
        ;    NEQ     NonZero (Not Equal)                     
        ; 

ifBNZE::                                ; NonZero (Not Equal)

        ; OBJ_00:       ifBNZE
        ; OBJ_01:       XX .SHL. 8 + YY

        bit             2, b            ; is Z = 0 ?

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_1 (Jump Address)

        ret             nz

        ld              sp, hl          ; sp = OBJ_1 (Jump Address)

        reti                            ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     GE     S xor O = 0                    1101     
        ;            Greater than or Equal                   
        ; 

ifBGE::                                 ; Greater than or Equal

        ; OBJ_00:       ifBGE
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        add             a, a            ; a = .1.SZOC.
        add             a, a            ;   = 1.SZOC.. (put O in bit #3)
        xor             b               ; put S xor O in bit #3

        pop             hl              ; hl <- OBJ_1 (Jump Address)

        and             $08             ; is S xor O = 0 ?
        ret             nz

        ld              sp, hl          ; sp = OBJ_1 (Jump Address)

        reti                            ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     GT     Z or ( S xor O ) = 0           1110     
        ;            Greater Than                            
        ; 

ifBGT::                                 ; Greater Than

        ; OBJ_00:       ifBGT
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        add             a, a            ; a = .1.SZOC.
        add             a, a            ;   = 1.SZOC.. (put O in bit #3)
        xor             b               ; put S xor O in bit #3
        rrca                            ;             in bit #2
        or              b               ; put Z or ( S xor O ) bit #2

        pop             hl              ; hl <- OBJ_1 (Jump Address)

        and             $04             ; is Z or ( S xor O ) = 0 ?
        ret             nz

        ld              sp, hl          ; sp = OBJ_1 (Jump Address)

        reti                            ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;    ESC     C xor S = 0                    1111     
        ;            Equal Sign and Carry                    
        ; 

ifBESC::                                ; Equal Sign and Carry

        ; OBJ_00:       ifBESC
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        swap            a               ;   = SZOC..1.
        rlca                            ;   = ZOC..1.S (put S in bit #0)
        xor             b               ; put C xor S in bit #0

        pop             hl              ; hl <- OBJ_1 (Jump Address)

        rrca                            ; is C xor S = 0 ? (put in CY)
        ret             c

        ld              sp, hl          ; sp = OBJ_1 (Jump Address)

        reti                            ; return, enabling interrupts

; ---------------------------------------------------------------------------
;              .::** Code for ifBcond (BREAK versions) **::.
; ---------------------------------------------------------------------------

ifDECR_R7::                             ; DECRement R7

        ; NOTE:  It's assumed that the flags (Status) would never be
        ;        tested as a result of an DECR R7, thus there exists
        ;        only one version of this instruction.

        ; OBJ_00:       ifDECR_R7

        di                              ; don't allow VBlankFunc to interrupt

        push            af              ; back up 1 program word (in 1 byte)

ifDECR_R7_1::                           ; entry point for ifHLT

        di                              ; don't allow VBlankFunc to interrupt

        jr              ifB_BREAK_2

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     C      C = 1                          0001     
        ;    LGT     Carry                                   
        ;            (Logical Greater Than)                  
        ; 

ifBC_BREAK::                            ; Carry (Logical Greater Than)

        ; OBJ_00:       ifBC_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        rrca                            ; is C = 1 ?
        jr              c, ifB_BREAK_1  ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     OV     O = 1                          0010     
        ;            Overflow                                
        ; 

ifBOV_BREAK::                           ; Overflow

        ; OBJ_00:       ifBOV_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        bit             1, b            ; is O = 1 ?

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        jr              nz, ifB_BREAK_1 ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

ifB_BREAK::                             ; Unconditional

        ; OBJ_00:       ifB_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl
ifB_BREAK_1::
        ld              sp, hl          ; sp <- OBJ_01 (Jump Address)
ifB_BREAK_2::
        ld              [R7], sp        ; update [R7]
        ld              a, b
        ld              [Status], a     ; update [Status]

        ld              sp, SP_0
        pop             hl
        ld              sp, hl          ; restore stack pointer

        reti                            ; return to Process's caller

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     PL     S = 0                          0011     
        ;            Plus                                    
        ; 

ifBPL_BREAK::                           ; Plus

        ; OBJ_00:       ifBPL_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        bit             3, b            ; is S = 0 ?

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        jr              z, ifB_BREAK_1  ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     ZE     Z = 1                          0100     
        ;     EQ     Zero (Equal)                            
        ; 

ifBZE_BREAK::                           ; Zero (Equal)

        ; OBJ_00:       ifBZE_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        bit             2, b            ; is Z = 1 ?

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        jr              nz, ifB_BREAK_1 ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     LT     S xor O = 1                    0101     
        ;            Less Than                               
        ; 

ifBLT_BREAK::                           ; Less Than

        ; OBJ_00:       ifBLT_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        add             a, a            ; a = .1.SZOC.
        add             a, a            ;   = 1.SZOC.. (put O in bit #3)
        xor             b               ; put S xor O in bit #3

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        and             $08             ; is S xor O = 1 ?
        jr              nz, ifB_BREAK_1 ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     LE     Z or ( S xor O ) = 1           0110     
        ;            Less than or Equal                      
        ; 

ifBLE_BREAK::                           ; Less than or Equal

        ; OBJ_00:       ifBLE_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        add             a, a            ; a = .1.SZOC.
        add             a, a            ;   = 1.SZOC.. (put O in bit #3)
        xor             b               ; put S xor O in bit #3
        rrca                            ;             in bit #2
        or              b               ; put Z or ( S xor O ) bit #2

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        and             $04             ; is Z or ( S xor O ) = 1 ?
        jr              nz, ifB_BREAK_1 ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;    USC     C xor S = 1                    0111     
        ;            Unequal Sign and Carry                  
        ; 

ifBUSC_BREAK::                          ; Unequal Sign and Carry

        ; OBJ_00:       ifBUSC_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        swap            a               ;   = SZOC..1.
        rlca                            ;   = ZOC..1.S (put S in bit #0)
        xor             b               ; put C xor S in bit #0

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        rrca                            ; is C xor S = 1 ? (put in CY)
        jr              c, ifB_BREAK_1  ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

ifNOPP_BREAK::                          ; No OPeration

        ; OBJ_00:       ifB_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     NC     C = 0                          1001     
        ;    LLT     No Carry                                
        ;            (Logical Less Than)                     
        ; 

ifBNC_BREAK::                           ; No Carry (Logical Less Than)

        ; OBJ_00:       ifBNC_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_1 (Jump Address)

        rrca                            ; is C = 0 ? (put in CY)
        jr              nc, ifB_BREAK_1 ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;    NOV     O = 0                          1010     
        ;            No Overflow                             
        ; 

ifBNOV_BREAK::                          ; No Overflow

        ; OBJ_00:       ifBNOV_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        bit             1, b            ; is O = 0 ?

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_1 (Jump Address)

        jr              z, ifB_BREAK_1  ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     MI     S = 1                          1011     
        ;            Minus                                   
        ; 

ifBMI_BREAK::                           ; Minus

        ; OBJ_00:       ifBMI_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        bit             3, b            ; is S = 1 ?

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_1 (Jump Address)

        jr              nz, ifB_BREAK_1 ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;    NZE     Z = 0                          1100     
        ;    NEQ     NonZero (Not Equal)                     
        ; 

ifBNZE_BREAK::                          ; NonZero (Not Equal)

        ; OBJ_00:       ifBNZE_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        bit             2, b            ; is Z = 0 ?

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_1 (Jump Address)

        jr              z, ifB_BREAK_1  ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     GE     S xor O = 0                    1101     
        ;            Greater than or Equal                   
        ; 

ifBGE_BREAK::                           ; Greater than or Equal

        ; OBJ_00:       ifBGE_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        add             a, a            ; a = .1.SZOC.
        add             a, a            ;   = 1.SZOC.. (put O in bit #3)
        xor             b               ; put S xor O in bit #3

        pop             hl              ; hl <- OBJ_01 (Jump Address)

        and             $08             ; is S xor O = 0 ?
        jr              z, ifB_BREAK_1  ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;     GT     Z or ( S xor O ) = 0           1110     
        ;            Greater Than                            
        ; 

ifBGT_BREAK::                           ; Greater Than

        ; OBJ_00:       ifBGT_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        add             a, a            ; a = .1.SZOC.
        add             a, a            ;   = 1.SZOC.. (put O in bit #3)
        xor             b               ; put S xor O in bit #3
        rrca                            ;             in bit #2
        or              b               ; put Z or ( S xor O ) bit #2

        pop             hl              ; hl <- OBJ_1 (Jump Address)

        and             $04             ; is Z or ( S xor O ) = 0 ?
        jr              z, ifB_BREAK_1  ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

        ; Ŀ
        ;                                        OBJECT CODE 
        ;  MNEMONIC  BRANCH CONDITION            DESIGNATION 
        ; Ĵ
        ;    ESC     C xor S = 0                    1111     
        ;            Equal Sign and Carry                    
        ; 

ifBESC_BREAK::                          ; Equal Sign and Carry

        ; OBJ_00:       ifBESC_BREAK
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, b            ; a = ..1.SZOC (Status)

        di                              ; don't allow VBlankFunc to interrupt

        swap            a               ;   = SZOC..1.
        rlca                            ;   = ZOC..1.S (put S in bit #0)
        xor             b               ; put C xor S in bit #0

        pop             hl              ; hl <- OBJ_1 (Jump Address)

        rrca                            ; is C xor S = 0 ? (put in CY)
        jr              nc, ifB_BREAK_1 ; yes, so break out of Process

        RETURN          0               ; return, enabling interrupts

; ---------------------------------------------------------------------------
;         .::** Code for MVO and MVO to SPLO, SPHI, and SRHI **::.
; ---------------------------------------------------------------------------

ifMVO_STIC::                            ; MoVe Out to STIC

        ; OBJ_00:       ifMVO_STIC_1
        ; OBJ_01:       BB .SHL. 8 + RS .SHL. 1

ifMVO_8_R0::                            ; MoVe Out to 8-bit w/ [R0]

        ; OBJ_00:       ifMVO_8_R0
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, [R0]         ; a <- LOW byte of [R0]

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl], a         ; store out LOW of [R0] to SCRATCHPAD

        RETURN          0               ; return, enabling interrupts

        ; ^ 14 cycles, 7 bytes

ifMVO_8_R1::                            ; MoVe Out to 8-bit w/ [R1]

        ; OBJ_00:       ifMVO_8_R1
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, [R1]         ; a <- LOW byte of [R1]

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl], a         ; store out LOW of [R1] to SCRATCHPAD

        RETURN          0               ; return, enabling interrupts

        ; ^ 14 cycles, 7 bytes

ifMVO_8_R2::                            ; MoVe Out to 8-bit w/ [R2]

        ; OBJ_00:       ifMVO_8_R2
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, [R2]         ; a <- LOW byte of [R2]

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl], a         ; store out LOW of [R2] to SCRATCHPAD

        RETURN          0               ; return, enabling interrupts

        ; ^ 14 cycles, 7 bytes

ifMVO_8_R3::                            ; MoVe Out to 8-bit w/ [R3]

        ; OBJ_00:       ifMVO_8_R3
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, [R3]         ; a <- LOW byte of [R3]

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl], a         ; store out LOW of [R3] to SCRATCHPAD

        RETURN          0               ; return, enabling interrupts

        ; ^ 14 cycles, 7 bytes

ifMVO_8_R4::                            ; MoVe Out to 8-bit w/ [R4]

        ; OBJ_00:       ifMVO_8_R4
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, [R4]         ; a <- LOW byte of [R4]

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl], a         ; store out LOW of [R4] to SCRATCHPAD

        RETURN          0               ; return, enabling interrupts

        ; ^ 14 cycles, 7 bytes

ifMVO_8_R5::                            ; MoVe Out to 8-bit w/ [R5]

        ; OBJ_00:       ifMVO_8_R5
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, [R5]         ; a <- LOW byte of [R5]

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl], a         ; store out LOW of [R5] to SCRATCHPAD

        RETURN          0               ; return, enabling interrupts

        ; ^ 14 cycles, 7 bytes

ifMVO_8_R6::                            ; MoVe Out to 8-bit w/ [R6]

        ; OBJ_00:       ifMVO_8_R6
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              a, [R6]         ; a <- LOW byte of [R6]

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl], a         ; store out LOW of [R6] to SCRATCHPAD

        RETURN          0               ; return, enabling interrupts

        ; ^ 14 cycles, 7 bytes

ifMVO_8_R7::                            ; MoVe Out to 8-bit w/ [R7]

        ; NOTE:         Since MVO R7, $XXYY is likely to be used only
        ;               for setting a flag to a non-zero value, moving
        ;               out R6 should be equally effective.

        ; OBJ_00:       ifMVO_8_R6
        ; OBJ_01:       XX .SHL. 8 + YY

ifMVO_AY8914::                          ; MoVe Out to AY8914

        ; OBJ_00:       ifMVO_AY8914_1
        ; OBJ_01:       BB .SHL. 9 + RS .SHL. 1

ifMVO_BKTB::                            ; MoVe Out to BacKTab

        ; OBJ_00:       ifMVO_BKTB_1
        ; OBJ_01:       BB .SHL. 8 + RS .SHL. 1

ifMVO_2F0h::                            ; MoVe Out to [02F0] (UDB)

        ; OBJ_00:       ifMVO_2F0h
        ; OBJ_01:       R0 + RS .SHL. 1

        ld              a, 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              [UDB_Flag], a   ; set UDB_Flag = 1

        pop             hl              ; hl <- OBJ_01 (RS)

        ld              a, [hl+]        ; a <- LOW byte of [RS]
        ld              d, [hl]         ; d <- HIGH byte of [RS]

        ld              hl, SYSTEM_RAM + $0F0 * 2

        ld              [hl+], a        ; store out LOW byte of [RS]
        ld              [hl], d         ; store out HIGH byte of [RS]

        RETURN          0               ; return, enabling interrupts

        ; ^ 25 cycles, 15 bytes

ifMVO_16_R0::                           ; MoVe Out to 16-bit w/ [R0]

        ; OBJ_00:       ifMVO_16_R0
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R0          ; hl -> R0

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [hl+]        ; a <- LOW byte of [R0]
        ld              d, [hl]         ; d <- HIGH byte of [R0]

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl+], a        ; store out LOW byte of [R0]
        ld              [hl], d         ; store out HIGH byte of [R0]

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVO_16_R1::                           ; MoVe Out to 16-bit w/ [R1]

        ; OBJ_00:       ifMVO_16_R1
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R1          ; hl -> R1

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [hl+]        ; a <- LOW byte of [R1]
        ld              d, [hl]         ; d <- HIGH byte of [R1]

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl+], a        ; store out LOW byte of [R1]
        ld              [hl], d         ; store out HIGH byte of [R1]

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVO_16_R2::                           ; MoVe Out to 16-bit w/ [R2]

        ; OBJ_00:       ifMVO_16_R2
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R2          ; hl -> R2

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [hl+]        ; a <- LOW byte of [R2]
        ld              d, [hl]         ; d <- HIGH byte of [R2]

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl+], a        ; store out LOW byte of [R2]
        ld              [hl], d         ; store out HIGH byte of [R2]

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVO_16_R3::                           ; MoVe Out to 16-bit w/ [R3]

        ; OBJ_00:       ifMVO_16_R3
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R3          ; hl -> R3

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [hl+]        ; a <- LOW byte of [R3]
        ld              d, [hl]         ; d <- HIGH byte of [R3]

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl+], a        ; store out LOW byte of [R3]
        ld              [hl], d         ; store out HIGH byte of [R3]

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVO_16_R4::                           ; MoVe Out to 16-bit w/ [R4]

        ; OBJ_00:       ifMVO_16_R4
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R4          ; hl -> R4

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [hl+]        ; a <- LOW byte of [R4]
        ld              d, [hl]         ; d <- HIGH byte of [R4]

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl+], a        ; store out LOW byte of [R4]
        ld              [hl], d         ; store out HIGH byte of [R4]

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVO_16_R5::                           ; MoVe Out to 16-bit w/ [R5]

        ; OBJ_00:       ifMVO_16_R5
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R5          ; hl -> R5

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [hl+]        ; a <- LOW byte of [R5]
        ld              d, [hl]         ; d <- HIGH byte of [R5]

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl+], a        ; store out LOW byte of [R5]
        ld              [hl], d         ; store out HIGH byte of [R5]

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVO_16_R6::                           ; MoVe Out to 16-bit w/ [R6]

        ; OBJ_00:       ifMVO_16_R6
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R6          ; hl -> R6

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [hl+]        ; a <- LOW byte of [R6]
        ld              d, [hl]         ; d <- HIGH byte of [R6]

        pop             hl              ; hl <- OBJ_01 (XXYY)

        ld              [hl+], a        ; store out LOW byte of [R6]
        ld              [hl], d         ; store out HIGH byte of [R6]

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVO_GALO::                            ; MoVe Out to GrAm LOw

        ; OBJ_00:       ifMVO_GALO_1
        ; OBJ_01:       BB .SHL. 8 + RS .SHL. 1

ifMVO_GAHI::                            ; MoVe Out to GrAm HIgh

        ; OBJ_00:       ifMVO_GAHI_1
        ; OBJ_01:       BB .SHL. 8 + RS .SHL. 1

; ---------------------------------------------------------------------------
;                  .::** MVO implied instructions **::.
; ---------------------------------------------------------------------------

ifMVO@_R13::                            ; MoVe Out [RS] to [[RM]]

        ; OBJ_00:       RSRM_GTA|ifMVO@_R13

        RPAIR_HEAD      hl, 1, 3, 5, 0, rtNone  ; l <- LOW byte of RM
                                                ; h <- LOW byte of RS

        di                              ; * don't allow VBlankFunc to interrupt

        ld              e, h
        ld              d, c            ; de -> RS
        ld              h, c            ; hl -> RM

        ld              a, [hl+]        ; a <- LOW byte of [RM]
        ld              h, [hl]         ; h <- HIGH byte of [RM]

        POKE                            ; [[RM]] <- [RS]

        ; ^ 3/6 + 10 = 13/16 cycles

ifMVO@_R45::                            ; MoVe Out [RS] to [[RM]]

        ; NOTE:  This routine will also handle MVO@ RS, R6 in the event
        ;        that the instruction is found to write to other regions
        ;        of memory besides SYSTEM_RAM.  Otherwise, it is handled
        ;        by the routines "ifMVO@_R6" and "ifMVO@_R5_R6" (see below).
        ;
        ;        If RS = R4, R5, R6 or R7, then RS = RM is not supported
        ;        (evidently not true, see Beauty & The Beast @ $5871).

        ; OBJ_00:       RSRM_GTA|ifMVO@_R45

        RPAIR_HEAD      hl, 6, 4, 6, 0, rtNone  ; l <- LOW byte of RM
                                                ; h <- LOW byte of RS

        di                              ; don't allow VBlankFunc to interrupt

        ld              e, h
        ld              d, c            ; de -> RS
        ld              h, c            ; hl -> RM

        ld              a, [hl]         ; a <- LOW byte of [RM]

        inc             [hl]            ; inc LOW byte of [RM]
        jr              z, ifMVO@_R45_HiInc

        inc             l               ; pos @ HIGH byte of [RM]
        ld              h, [hl]         ; h <- HIGH byte of [RM]

        POKE                            ; [[RM]] <- [RS]

        ; ^ 3/6 + 16 = 19/22 cycles

ifMVO@_R45_HiInc:

        inc             l               ; pos @ HIGH byte of [RM]
        inc             [hl]            ; inc HIGH byte of [RM]

        ld              h, [hl]
        dec             h               ; h <- HIGH byte of [RM]

        POKE                            ; [[RM]] <- [RS]

ifMVO@_R0_R5::                          ; MoVe Out [R0] to [[R5]]

        ; NOTE:  This is a special, optimized version of the instruction
        ;        MVO@ R0, R5, since it occurs VERY frequently.  It only
        ;        handles the one combination of registers.

        ; OBJ_00:       ifMVO@_R0_R5

        ld              de, R0          ; de -> R0
        di                              ; don't allow VBlankFunc to interrupt
        ld              hl, R5          ; hl -> R5

        ld              a, [hl]         ; a <- LOW byte of [R5]

        inc             [hl]            ; inc LOW byte of [R5]
        jr              z, ifMVO@_R45_HiInc

        inc             l               ; pos @ HIGH byte of [R5]
        ld              h, [hl]         ; h <- HIGH byte of [R5]

        POKE                            ; [[R5]] <- [R0]

        ; ^ 19 cycles

; ---------------------------------------------------------------------------
;                      .::** Code for MVO@ RS, R6 **::.
; ---------------------------------------------------------------------------

ifMVO@_R0_R6::                          ; MoVe Out [R0] to [[R6]]

        ; OBJ_00:       ifMVO@_R0_R6

        ld              hl, R0          ; hl -> R0

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [hl+]        ; a <- LOW byte of [RS]
        ld              d, [hl]         ; d <- HIGH byte of [RS]

        ld              l, R6 & $FF     ; hl -> R6

        inc             [hl]            ; inc LOW byte of [R6]

        ld              l, [hl]         ; l <- LOW byte of [R6]

        ld              [hl-], a        ; store out LOW byte of [RS]
        inc             h
        ld              [hl], d         ; store out HIGH byte of [RS]

        RETURN          0               ; return, enabling interrupts

        ; ^ 20 + 4 = 24 cycles, 14 bytes

ifMVO@_R1_R6::                          ; MoVe Out [R1] to [[R6]]

        ; OBJ_00:       ifMVO@_R1_R6

        ld              hl, R1          ; hl -> R1

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [hl+]        ; a <- LOW byte of [RS]
        ld              d, [hl]         ; d <- HIGH byte of [RS]

        ld              l, R6 & $FF     ; hl -> R6

        inc             [hl]            ; inc LOW byte of [R6]

        ld              l, [hl]         ; l <- LOW byte of [R6]

        ld              [hl-], a        ; store out LOW byte of [RS]
        inc             h
        ld              [hl], d         ; store out HIGH byte of [RS]

        RETURN          0               ; return, enabling interrupts

        ; ^ 20 + 4 = 24 cycles, 14 bytes

ifMVO@_R2_R6::                          ; MoVe Out [R2] to [[R6]]

        ; OBJ_00:       ifMVO@_R2_R6

        ld              hl, R2          ; hl -> R2

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [hl+]        ; a <- LOW byte of [RS]
        ld              d, [hl]         ; d <- HIGH byte of [RS]

        ld              l, R6 & $FF     ; hl -> R6

        inc             [hl]            ; inc LOW byte of [R6]

        ld              l, [hl]         ; l <- LOW byte of [R6]

        ld              [hl-], a        ; store out LOW byte of [RS]
        inc             h
        ld              [hl], d         ; store out HIGH byte of [RS]

        RETURN          0               ; return, enabling interrupts

        ; ^ 20 + 4 = 24 cycles, 14 bytes

ifMVO@_R3_R6::                          ; MoVe Out [R3] to [[R6]]

        ; OBJ_00:       ifMVO@_R3_R6

        ld              hl, R3          ; hl -> R3

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [hl+]        ; a <- LOW byte of [RS]
        ld              d, [hl]         ; d <- HIGH byte of [RS]

        ld              l, R6 & $FF     ; hl -> R6

        inc             [hl]            ; inc LOW byte of [R6]

        ld              l, [hl]         ; l <- LOW byte of [R6]

        ld              [hl-], a        ; store out LOW byte of [RS]
        inc             h
        ld              [hl], d         ; store out HIGH byte of [RS]

        RETURN          0               ; return, enabling interrupts

        ; ^ 20 + 4 = 24 cycles, 14 bytes

ifMVO@_R4_R6::                          ; MoVe Out [R4] to [[R6]]

        ; OBJ_00:       ifMVO@_R4_R6

        ld              hl, R4          ; hl -> R4

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [hl+]        ; a <- LOW byte of [RS]
        ld              d, [hl]         ; d <- HIGH byte of [RS]

        ld              l, R6 & $FF     ; hl -> R6

        inc             [hl]            ; inc LOW byte of [R6]

        ld              l, [hl]         ; l <- LOW byte of [R6]

        ld              [hl-], a        ; store out LOW byte of [RS]
        inc             h
        ld              [hl], d         ; store out HIGH byte of [RS]

        RETURN          0               ; return, enabling interrupts

        ; ^ 20 + 4 = 24 cycles, 14 bytes

ifMVO@_R5_R6::                          ; MoVe Out [R5] to [[R6]]

        ; OBJ_00:       ifMVO@_R5_R6

        ld              hl, R5          ; hl -> R5

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [hl+]        ; a <- LOW byte of [R5]
        ld              d, [hl]         ; d <- HIGH byte of [R5]

        inc             l               ; hl -> R6

        inc             [hl]            ; inc LOW byte of [R6]

        ld              l, [hl]         ; l <- LOW byte of [R6]

        ld              [hl-], a        ; store out LOW byte of [R5]
        inc             h
        ld              [hl], d         ; store out HIGH byte of [R5]

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 + 4 = 23 cycles, 13 bytes

; ---------------------------------------------------------------------------
;         .::** Code for MVI from SPLO, SPHI, and SRHI **::.
; ---------------------------------------------------------------------------

ifMVI_8_R0::                            ; MoVe In from 8-bit region to [R0]

        ; OBJ_00:       ifMVI_8_R0
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R0          ; hl -> R0

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; de -> SCRATCHPAD_RAM

        ld              a, [de]         ; a <- 8-bit value from SCRATCHPAD

        ld              [hl+], a        ; store it out to LOW byte of [RD]
        ld              [hl], 0         ; set HIGH byte of [RD] = 0

        RETURN          0               ; return, enabling interrupts

        ; ^ 18 cycles, 10 bytes

ifMVI_8_R1::                            ; MoVe In from 8-bit region to [R1]

        ; OBJ_00:       ifMVI_8_R1
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R1          ; hl -> R1

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; de -> SCRATCHPAD_RAM

        ld              a, [de]         ; a <- 8-bit value from SCRATCHPAD

        ld              [hl+], a        ; store it out to LOW byte of [RD]
        ld              [hl], 0         ; set HIGH byte of [RD] = 0

        RETURN          0               ; return, enabling interrupts

        ; ^ 18 cycles, 10 bytes

ifMVI_8_R2::                            ; MoVe In from 8-bit region to [R2]

        ; OBJ_00:       ifMVI_8_R2
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R2          ; hl -> R2

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; de -> SCRATCHPAD_RAM

        ld              a, [de]         ; a <- 8-bit value from SCRATCHPAD

        ld              [hl+], a        ; store it out to LOW byte of [RD]
        ld              [hl], 0         ; set HIGH byte of [RD] = 0

        RETURN          0               ; return, enabling interrupts

        ; ^ 18 cycles, 10 bytes

ifMVI_8_R3::                            ; MoVe In from 8-bit region to [R3]

        ; OBJ_00:       ifMVI_8_R3
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R3          ; hl -> R0

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; de -> SCRATCHPAD_RAM

        ld              a, [de]         ; a <- 8-bit value from SCRATCHPAD

        ld              [hl+], a        ; store it out to LOW byte of [RD]
        ld              [hl], 0         ; set HIGH byte of [RD] = 0

        RETURN          0               ; return, enabling interrupts

        ; ^ 18 cycles, 10 bytes

ifMVI_8_R4::                            ; MoVe In from 8-bit region to [R4]

        ; OBJ_00:       ifMVI_8_R4
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R4          ; hl -> R4

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; de -> SCRATCHPAD_RAM

        ld              a, [de]         ; a <- 8-bit value from SCRATCHPAD

        ld              [hl+], a        ; store it out to LOW byte of [RD]
        ld              [hl], 0         ; set HIGH byte of [RD] = 0

        RETURN          0               ; return, enabling interrupts

        ; ^ 18 cycles, 10 bytes

ifMVI_8_R5::                            ; MoVe In from 8-bit region to [R5]

        ; OBJ_00:       ifMVI_8_R5
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R5          ; hl -> R5

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; de -> SCRATCHPAD_RAM

        ld              a, [de]         ; a <- 8-bit value from SCRATCHPAD

        ld              [hl+], a        ; store it out to LOW byte of [RD]
        ld              [hl], 0         ; set HIGH byte of [RD] = 0

        RETURN          0               ; return, enabling interrupts

        ; ^ 18 cycles, 10 bytes

ifMVI_8_R6::                            ; MoVe In from 8-bit region to [R6]

        ; OBJ_00:       ifMVI_8_R6
        ; OBJ_01:       XX .SHL. 8 + YY

        ld              hl, R6          ; hl -> R6

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; de -> SCRATCHPAD_RAM

        ld              a, [de]         ; a <- 8-bit value from SCRATCHPAD

        ld              [hl+], a        ; store it out to LOW byte of [RD]
        ld              [hl], 0         ; set HIGH byte of [RD] = 0

        RETURN          0               ; return, enabling interrupts

        ; ^ 18 cycles, 10 bytes

ifMVI_16_R0::                           ; MoVe In from 16-bit region to [R0]

        ; OBJ_00:       ifMVI_16_R0
        ; OBJ_01:       XX .SHL. 8 + YY

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl -> SYSTEM_RAM

        ld              a, [hl+]        ; a <- LOW byte of 16-bit value
        ld              d, [hl]         ; d <- HIGH byte of 16-bit value

        ld              hl, R0          ; hl -> R0

        ld              [hl+], a        ; store out LOW byte of 16-bit value
        ld              [hl], d         ; store out HIGH byte of 16-bit value

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVI_16_R1::                           ; MoVe In from 16-bit region to [R1]

        ; OBJ_00:       ifMVI_16_R1
        ; OBJ_01:       XX .SHL. 8 + YY

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl -> SYSTEM_RAM

        ld              a, [hl+]        ; a <- LOW byte of 16-bit value
        ld              d, [hl]         ; d <- HIGH byte of 16-bit value

        ld              hl, R1          ; hl -> R1

        ld              [hl+], a        ; store out LOW byte of 16-bit value
        ld              [hl], d         ; store out HIGH byte of 16-bit value

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVI_16_R2::                           ; MoVe In from 16-bit region to [R2]

        ; OBJ_00:       ifMVI_16_R2
        ; OBJ_01:       XX .SHL. 8 + YY

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl -> SYSTEM_RAM

        ld              a, [hl+]        ; a <- LOW byte of 16-bit value
        ld              d, [hl]         ; d <- HIGH byte of 16-bit value

        ld              hl, R2          ; hl -> R2

        ld              [hl+], a        ; store out LOW byte of 16-bit value
        ld              [hl], d         ; store out HIGH byte of 16-bit value

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVI_16_R3::                           ; MoVe In from 16-bit region to [R3]

        ; OBJ_00:       ifMVI_16_R3
        ; OBJ_01:       XX .SHL. 8 + YY

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl -> SYSTEM_RAM

        ld              a, [hl+]        ; a <- LOW byte of 16-bit value
        ld              d, [hl]         ; d <- HIGH byte of 16-bit value

        ld              hl, R3          ; hl -> R3

        ld              [hl+], a        ; store out LOW byte of 16-bit value
        ld              [hl], d         ; store out HIGH byte of 16-bit value

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVI_16_R4::                           ; MoVe In from 16-bit region to [R4]

        ; OBJ_00:       ifMVI_16_R4
        ; OBJ_01:       XX .SHL. 8 + YY

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl -> SYSTEM_RAM

        ld              a, [hl+]        ; a <- LOW byte of 16-bit value
        ld              d, [hl]         ; d <- HIGH byte of 16-bit value

        ld              hl, R4          ; hl -> R4

        ld              [hl+], a        ; store out LOW byte of 16-bit value
        ld              [hl], d         ; store out HIGH byte of 16-bit value

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVI_16_R5::                           ; MoVe In from 16-bit region to [R5]

        ; OBJ_00:       ifMVI_16_R5
        ; OBJ_01:       XX .SHL. 8 + YY

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl -> SYSTEM_RAM

        ld              a, [hl+]        ; a <- LOW byte of 16-bit value
        ld              d, [hl]         ; d <- HIGH byte of 16-bit value

        ld              hl, R5          ; hl -> R5

        ld              [hl+], a        ; store out LOW byte of 16-bit value
        ld              [hl], d         ; store out HIGH byte of 16-bit value

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVI_16_R6::                           ; MoVe In from 16-bit region to [R6]

        ; OBJ_00:       ifMVI_16_R6
        ; OBJ_01:       XX .SHL. 8 + YY

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl -> SYSTEM_RAM

        ld              a, [hl+]        ; a <- LOW byte of 16-bit value
        ld              d, [hl]         ; d <- HIGH byte of 16-bit value

        ld              hl, R6          ; hl -> R6

        ld              [hl+], a        ; store out LOW byte of 16-bit value
        ld              [hl], d         ; store out HIGH byte of 16-bit value

        RETURN          0               ; return, enabling interrupts

        ; ^ 19 cycles, 10 bytes

ifMVI_R7::                              ; MoVe In (from any region) to [R7]

        ; OBJ_00:       ifMVI_R7
        ; OBJ_01:       XX .SHL. 8 + YY + 1

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl <- ADDR (real)

        ld              a, [hl-]        ; a <- HIGH byte of 16-bit value
        ld              e, [hl]         ; e <- LOW byte of 16-bit value

        jp              Jump_Op         ; conv. addr., switch bank if needed

ifMVI_21h::                             ; MoVe In from [21h] (set ColorStack)

        ; OBJ_00:       ifMVI_21h
        ; OBJ_01:       R0 + RD .SHL. 1

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl -> RD (should we modify it?)

ifMVI@_R13_STIC_21H::                   ; MoVe In implied from [21h] (set CS)

        ; OBJ_00:       ifMVI@_R13_STIC_21H

        xor             a               ; a = 0 (ColorStack)
        ld              [STIC_Mode_New], a

        ld              hl, BW_BW + 1   ; AP_BW + 1

        ld              a, BktWrite_CS & $FF
        ld              [hl+], a
        ld              [hl], BktWrite_CS >> 8

        RETURN          0               ; return, enabling interrupts

ifADD_21h::                             ; ADD to [21h] (set ColorStack)

        ; OBJ_00:       ifMVI_21h
        ; OBJ_01:       R0 + RD .SHL. 1

ifSUB_21h::                             ; SUB from [21h] (set ColorStack)

        ; OBJ_00:       ifMVI_21h
        ; OBJ_01:       R0 + RD .SHL. 1

ifCMP_21h::                             ; CMP [21h] (set ColorStack)

        ; OBJ_00:       ifMVI_21h
        ; OBJ_01:       R0 + RD .SHL. 1

ifAND_21h::                             ; AND [21h] (set ColorStack)

        ; OBJ_00:       ifMVI_21h
        ; OBJ_01:       R0 + RD .SHL. 1

ifXOR_21h::                             ; XOR [21h] (set ColorStack)

        ; OBJ_00:       ifMVI_21h
        ; OBJ_01:       R0 + RD .SHL. 1

; ---------------------------------------------------------------------------
;        .::** Code for MVI@ R1-R3, RD with and without SDBD **::.
; ---------------------------------------------------------------------------

        IF              R13_SDBD == 1

ifMVI@_R13_SDBD::                       ; MoVe In implied from R1-R3, SDBD

        ; OBJ_FF:       ifMVI@_R13_SDBD
        ; OBJ_00:       RD .SHL. 9 + RM .SHL. 1 + 1

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; l <- LOW byte of RM + 1
                                        ; h <- LOW byte of RD

        ld              e, h            ; save LOW byte of RD in e
        ld              h, c            ; hl -> RM + 1

        ld              a, [hl-]
        ld              d, [hl]         ; ad <- [RM]

        PEEK            mpSDBD_R13      ; a <- [[RM]]

        ld              l, e
        ld              h, c            ; hl -> RD

        ld              [hl+], a        ; store out LOW byte of [RD]
        ld              [hl], a         ; store out HIGH byte of [RD]

        RETURN          mpROMB0         ; return, enabling interrupts

        ENDC

ifMVI@_R13_NoSDBD::                     ; MoVe In implied from R1-R3, NoSDBD

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RMRD_GTA|ifMVI@_R13_NoSDBD

        RPAIR_HEAD      hl, 3, 1, 5, 0, rtRM_P1  ; l <- LOW byte of RM + 1
                                                 ; h <- LOW byte of RD

        di                              ; needed in case RM = RD !!

        ld              e, h            ; save LOW byte of RD in e
        ld              h, c            ; hl -> RM + 1

        ld              a, [hl-]
        ld              d, [hl]         ; ad <- [RM]

        PEEK            0, d            ; da <- [[RM]]

        ld              l, e
        ld              h, c            ; hl -> RD

        ld              [hl+], a        ; store out LOW byte of [RD]
        ld              [hl], d         ; store out HIGH byte of [RD]

        RETURN          mpROMB0         ; return, enabling interrupts

        ; ^ 3/6 + 7 + 15 + 6 + 4 = 35/38 cycles

ifMVI@_R13_NoSDBD_NBS::                 ; MoVe In implied from R1-R3, NoSDBD

        ; NOTE:  This routine does not need to disable interrupts,
        ;        since it can safely be interrupted at any time,
        ;        and then re-executed.

        ; OBJ_00:       RMRD_GTA|ifMVI@_R13_NoSDBD_NBS

        RPAIR_HEAD      hl, 3, 1, 5, 0, rtNone  ; l <- LOW byte of RM
                                                ; h <- LOW byte of RD

        di                              ; needed in case RM = RD !!

        ld              e, h            ; save LOW byte of RD in e
        ld              h, c            ; hl -> RM

        ld              a, [hl+]
        ld              l, [hl]         ; la <- [RM]

        inc             h               ; hl -> AddrHiToReal

        ld              h, [hl]         ; h <- HIGH of PeekAddr (Real) >> 1
        ld              l, a

        add             hl, hl          ; hl = PeekAddr (Real)

        ld              a, [hl+]        ; a <- LOW byte of 16-bit value
        ld              d, [hl]         ; d <- HIGH byte of 16-bit value

        ld              l, e
        ld              h, c            ; hl -> RD

        ld              [hl+], a        ; store out LOW byte of [RD]
        ld              [hl], d         ; store out HIGH byte of [RD]

        RETURN          0               ; return, enabling interrupts

        ; ^ 3/6 + 23 + 4 = 30/33 cycles vs. 39/42 cycles (saves 9 cycles)

ifMVI@_R13_R7::                         ; MoVe In implied from R1-R3 to [R7]

        ; NOTE:  It is assumed an SDBD would _never_ occur with this
        ;        instruction, since you can only retrieve addresses
        ;        in which the LOW and HIGH bytes are identical, which
        ;        is ridiculously unlikely.

        ; OBJ_00:       RM_GTA|ifMVI@_R13_R7

        RSING_TAB       l, 3, 1, rtRM_P1  ; l <- LOW byte of RM + 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RM + 1

        ld              a, [hl-]
        ld              d, [hl]         ; ad <- [RM]

        PEEK            0, l            ; la <- [[RM]]

        jp              Jump_Peek       ; convert addr, switch bank if needed

; ---------------------------------------------------------------------------
;        .::** Code for MVI@ R4-R5, RD with and without SDBD **::.
; ---------------------------------------------------------------------------

ifMVI@_R5_NoSDBD_STACK::                ; MoVe In [[R5]] to [R4], NoSDBD

        ; OBJ_00:       RD_GTA|ifMVI@_R5_NoSDBD_STACK

        RSING_TAB       e, 5, 0, rtNone  ; e <- LOW byte of RD

        di                              ; don't allow VBlankFunc to interrupt

        ld              hl, R5          ; hl -> R5

        inc             [hl]            ; inc LOW byte of [R5]

        ld              l, [hl]         ; l <- LOW byte of [R5] + 1

        ld              a, [hl-]        ; a <- LOW byte of [[R5]]
        inc             h
        ld              d, [hl]         ; d <- HIGH byte of [[R5]]

        ld              l, e
        ld              h, c            ; hl -> RD

        ld              [hl+], a        ; store out LOW byte of [R4]
        ld              [hl], d         ; store out HIGH byte of [R4]

        RETURN          0               ; return, enabling interrupts

ifMVI@_R45_NoSDBD_NBS::                 ; MoVe In implied from R4-R5, NoSDBD

        ; OBJ_00:       RMRD_GTA|ifMVI@_R45_NoSDBD_NBS

        RPAIR_HEAD      hl, 5, 4, 6, 0, rtRM_P1  ; l <- LOW byte of RM + 1
                                                 ; h <- HIGH byte of RD

        di                              ; don't allow VBlankFunc to interrupt

        ld              e, h            ; save LOW byte of RD in e
        ld              h, c            ; hl -> RM + 1 ( @ HIGH byte )

        ld              a, [hl-]
        ld              d, [hl]         ; ad <- [RM]

        inc             [hl]            ; inc LOW byte of [RM]
        jr              z, .HiInc
.Cont
        ld              l, a
        inc             h               ; pos @ AddrHiToReal

        ld              h, [hl]         ; h <- HIGH of PeekAddr (Real) >> 1
        ld              l, d

        add             hl, hl          ; hl = PeekAddr (Real)

        ld              a, [hl+]        ; a <- LOW byte of 16-bit value
        ld              d, [hl]         ; d <- HIGH byte of 16-bit value

        ld              l, e
        ld              h, c            ; hl -> RD

        ld              [hl+], a        ; store out to LOW byte of [RD]
        ld              [hl], d         ; store out to HIGH byte of [RD]

        RETURN          0               ; return, enabling interrupts

        ; ^ 3/6 + 29 + 4 = 36/39 cycles vs. 44/47 cycles (saves 8 cycles)

.HiInc

        inc             l               ; pos @ HIGH byte of [RM]
        inc             [hl]            ; inc HIGH byte of [RM]

        jr              .Cont

ifMVI@_R45::                            ; MoVe In implied from R4-R5

        ; OBJ_00:       RMRD_GTA|ifMVI@_R45

        RPAIR_HEAD      hl, 5, 4, 5, 0, rtRM_P1  ; l <- LOW byte of RM + 1
                                                 ; h <- HIGH byte of RD

        di                              ; don't allow VBlankFunc to interrupt

        ld              a, [SDBD]
        sub             c                        ; is SDBD = $D3 ?
        jr              nz, ifMVI@_R45_NoSDBD_1  ; no, so run non-SDBD rt.

        ld              [SDBD], a       ; clear SDBD

        ld              a, 2
        dec             l               ; l = LOW byte of RM

        jr              ifMVI@_R45_SDBD_1

ifMVI@_R45_NoSDBD::                     ; MoVe In implied from R4-R5, NoSDBD

        ; OBJ_00:       RMRD_GTA|ifMVI@_R45_NoSDBD

        RPAIR_HEAD      hl, 5, 4, 5, 0, rtRM_P1  ; l <- LOW byte of RM + 1
                                                 ; h <- HIGH byte of RD

        di                              ; don't allow VBlankFunc to interrupt

ifMVI@_R45_NoSDBD_1:                    ; entry point for ifMVI@_R45 (above)

        ld              e, h            ; save LOW byte of RD in e
        ld              h, c            ; hl -> RM + 1

        ld              a, [hl-]        ; a <- HIGH byte of [RM]
        ld              d, [hl]         ; d <- LOW byte of [RM]

        inc             [hl]            ; inc LOW byte of [RM]
        jr              z, .HiInc
.Cont
        PEEK            0, d            ; da <- [[RM]]

        ld              l, e
        ld              h, c            ; hl -> RD (in Echo RAM)

        ld              [hl+], a        ; store out to LOW byte of [RD]
        ld              [hl], d         ; store out to HIGH byte of [RD]

        RETURN          mpROMB0         ; return, enabling interrupts

        ; ^ 3/6 + 12 + 17 + 6 + 4 = 42/45 cycles

.HiInc

        inc             l               ; pos @ HIGH byte of [RM]
        inc             [hl]            ; inc HIGH byte of [RM]

        jr              .Cont

ifMVI@_R45_SDBD::                       ; MoVe In implied from R4-R5, SDBD

        ; NOTE:  The most that could be saved by coding separate versions
        ;        of this routine for R4 and R5 is 1 cycle, and since this
        ;        is much less common than the non-SDBD version, is probably
        ;        not worth the extra code space.

        ; OBJ_FF:       ifMVI@_R45_SDBD
        ; OBJ_00:       RD .SHL. 9 + RM .SHL. 1

        ld              a, 2            ; prepare for auto-increment of [RM]

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; l <- LOW byte of OBJ_1 (RM)
                                        ; h <- HIGH byte of OBJ_1 (RD)

ifMVI@_R45_SDBD_1:                      ; entry point for ifMVI@_R45 (above)

        ld              e, h            ; save LOW byte of RD in e
        ld              h, c            ; hl -> RM

        ld              d, [hl]         ; d <- LOW byte of [RM]
        add             d               ; a = LOW byte of [RM] + 2
        ld              [hl+], a        ; store out LOW byte of [RM] + 2

        ld              a, [hl]         ; * a <- HIGH byte of [RM]
        jr              c, .HiInc
.Cont
        PEEK            mpSDBD, d       ; da <- [[RM]]

        ld              l, e
        ld              h, c            ; hl -> RD

        ld              [hl+], a        ; store out to LOW byte of [RD]
        ld              [hl], d         ; store out to HIGH byte of [RD]

        RETURN          mpROMB0         ; return, enabling interrupts

        ; ^ 17 + 17 + 6 + 8 = 48 cycles

.HiInc

        inc             [hl]            ; inc HIGH byte of [RM]

        jr              .Cont

ifMVI@_R45_R7_SDBD::                    ; MoVe In implied R4-R5 to [R7]

        ; OBJ_FF:       ifMVI@_R45_R7_SDBD
        ; OBJ_00:       R0 + RM .SHL. 1

        ld              a, 2            ; prepare for auto-increment of [RM]

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; hl -> RM

        ld              d, [hl]         ; d <- LOW byte of [RM]
        add             d               ; a = LOW byte of [RM] + 2
        ld              [hl+], a        ; store out LOW byte of [RM] + 2

        ld              a, [hl]         ; a <- HIGH byte of [RM]
        jr              c, ifMVI@_R45_R7_SDBD_HiInc

ifMVI@_R45_R7_SDBD_Peek:

        PEEK            mpSDBD, l       ; la <- [[RM]]

Jump_Peek:

        JUMP            mpJAinLA        ; convert addr, switch bank if needed

ifMVI@_R45_R7_SDBD_HiInc:

        inc             [hl]            ; inc HIGH byte of [RM]

        jr              ifMVI@_R45_R7_SDBD_Peek

ifMVI@_R45_R7_NoSDBD::                  ; MoVe In implied from R4-R5 to [R7]

        ; OBJ_00:       RM_GTA|ifMVI@_R45_R7_NoSDBD

        RSING_TAB       l, 5, 4, rtRM_P1  ; l <- LOW byte of RM + 1

        di                              ; don't allow VBlankFunc to interrupt

        ld              h, c            ; hl -> RM + 1

        ld              a, [hl-]        ; a <- HIGH byte of [RM]
        ld              d, [hl]         ; d <- LOW byte of [RM]

        inc             [hl]            ; inc LOW byte of [RM]
        jr              z, .HiInc
.Cont
        PEEK            0, l            ; la <- [[RM]]

        jr              Jump_Peek       ; convert addr, switch bank if needed

.HiInc

        inc             l               ; pos @ HIGH byte of [RM]
        inc             [hl]            ; increment HIGH byte of [RM]

        jr              .Cont

; ---------------------------------------------------------------------------
;           .::** Code for MVI@ R6, RD (no SDBD is possible) **::.
; ---------------------------------------------------------------------------

ifMVI@_R6_R0::                          ; MoVe In implied from R6 to [R0]

        ; OBJ_00:       ifMVI@_R6_R0

        ld              hl, R6          ; hl -> R6

        di                              ; don't allow VBlankFunc to interrupt

        dec             [hl]            ; [R6] = [R6] - 1

        ld              l, [hl]         ; l <- LOW byte of [R6] - 1
        inc             h

        ld              a, [hl+]        ; a <- HIGH byte of [[R6] - 1]
        dec             h
        ld              d, [hl]         ; d <- LOW byte of [[R6] - 1]

        ld              l, R0 & $FF + 1 ; hl -> R0 + 1 ( @ HIGH byte )

        ld              [hl-], a        ; store out HIGH byte of [[R6] - 1]
        ld              [hl], d         ; store out LOW byte of [[R6] - 1]

        RETURN          0               ; return, enabling interrupts

        ; ^ 21 + 4 = 25 cycles, 15 bytes

ifMVI@_R6_R1::                          ; MoVe In implied from R6 to [R1]

        ; OBJ_00:       ifMVI@_R6_R1

        ld              hl, R6          ; hl -> R6

        di                              ; don't allow VBlankFunc to interrupt

        dec             [hl]            ; [R6] = [R6] - 1

        ld              l, [hl]         ; l <- LOW byte of [R6] - 1
        inc             h

        ld              a, [hl+]        ; a <- HIGH byte of [[R6] - 1]
        dec             h
        ld              d, [hl]         ; d <- LOW byte of [[R6] - 1]

        ld              l, R1 & $FF + 1 ; hl -> R1 + 1 ( @ HIGH byte )

        ld              [hl-], a        ; store out HIGH byte of [[R6] - 1]
        ld              [hl], d         ; store out LOW byte of [[R6] - 1]

        RETURN          0               ; return, enabling interrupts

        ; ^ 21 + 4 = 25 cycles, 15 bytes

ifMVI@_R6_R2::                          ; MoVe In implied from R6 to [R2]

        ; OBJ_00:       ifMVI@_R6_R2

        ld              hl, R6          ; hl -> R6

        di                              ; don't allow VBlankFunc to interrupt

        dec             [hl]            ; [R6] = [R6] - 1

        ld              l, [hl]         ; l <- LOW byte of [R6] - 1
        inc             h

        ld              a, [hl+]        ; a <- HIGH byte of [[R6] - 1]
        dec             h
        ld              d, [hl]         ; d <- LOW byte of [[R6] - 1]

        ld              l, R2 & $FF + 1 ; hl -> R2 + 1 ( @ HIGH byte )

        ld              [hl-], a        ; store out HIGH byte of [[R6] - 1]
        ld              [hl], d         ; store out LOW byte of [[R6] - 1]

        RETURN          0               ; return, enabling interrupts

        ; ^ 21 + 4 = 25 cycles, 15 bytes

ifMVI@_R6_R3::                          ; MoVe In implied from R6 to [R3]

        ; OBJ_00:       ifMVI@_R6_R3

        ld              hl, R6          ; hl -> R6

        di                              ; don't allow VBlankFunc to interrupt

        dec             [hl]            ; [R6] = [R6] - 1

        ld              l, [hl]         ; l <- LOW byte of [R6] - 1
        inc             h

        ld              a, [hl+]        ; a <- HIGH byte of [[R6] - 1]
        dec             h
        ld              d, [hl]         ; d <- LOW byte of [[R6] - 1]

        ld              l, R3 & $FF + 1 ; hl -> R3 + 1 ( @ HIGH byte )

        ld              [hl-], a        ; store out HIGH byte of [[R6] - 1]
        ld              [hl], d         ; store out LOW byte of [[R6] - 1]

        RETURN          0               ; return, enabling interrupts

        ; ^ 21 + 4 = 25 cycles, 15 bytes

ifMVI@_R6_R4::                          ; MoVe In implied from R6 to [R4]

        ; OBJ_00:       ifMVI@_R6_R4

        ld              hl, R6          ; hl -> R6

        di                              ; don't allow VBlankFunc to interrupt

        dec             [hl]            ; [R6] = [R6] - 1

        ld              l, [hl]         ; l <- LOW byte of [R6] - 1
        inc             h

        ld              a, [hl+]        ; a <- HIGH byte of [[R6] - 1]
        dec             h
        ld              d, [hl]         ; d <- LOW byte of [[R6] - 1]

        ld              l, R4 & $FF + 1 ; hl -> R4 + 1 ( @ HIGH byte )

        ld              [hl-], a        ; store out HIGH byte of [[R6] - 1]
        ld              [hl], d         ; store out LOW byte of [[R6] - 1]

        RETURN          0               ; return, enabling interrupts

        ; ^ 21 + 4 = 25 cycles, 15 bytes

ifMVI@_R6_R5::                          ; MoVe In implied from R6 to [R5]

        ; OBJ_00:       ifMVI@_R6_R5

        ld              hl, R6          ; hl -> R6

        di                              ; don't allow VBlankFunc to interrupt

        dec             [hl]            ; [R6] = [R6] - 1

        ld              l, [hl]         ; l <- LOW byte of [R6] - 1
        inc             h

        ld              a, [hl+]        ; a <- HIGH byte of [[R6] - 1]
        dec             h
        ld              d, [hl]         ; d <- LOW byte of [[R6] - 1]

        ld              l, R5 & $FF + 1 ; hl -> R5 + 1 ( @ HIGH byte )

        ld              [hl-], a        ; store out HIGH byte of [[R6] - 1]
        ld              [hl], d         ; store out LOW byte of [[R6] - 1]

        RETURN          0               ; return, enabling interrupts

        ; ^ 21 + 4 = 25 cycles, 15 bytes

ifMVI@_R6_R7::                          ; Move In implied from R6 to [R7]

        ; OBJ_00:       ifMVI@_R6_R7

        ld              hl, R6          ; hl -> R6

        di                              ; don't allow VBlankFunc to interrupt

        dec             [hl]            ; [R6] = [R6] - 1

        ld              l, [hl]         ; l <- LOW byte of [R6] - 1
        inc             h

        ld              a, [hl+]        ; a <- HIGH byte of [[R6] - 1]
        dec             h
        ld              e, [hl]         ; e <- LOW byte of [[R6] - 1]

        JUMP            mpR0inH | mpJAinAE

ifMVI@_R6_R7_NBS::                      ; Move In implied from R6 to [R7]

        ; OBJ_00:       ifMVI@_R6_R7_NBS

        ld              hl, R6          ; hl -> R6

        di                              ; don't allow VBlankFunc to interrupt

        dec             [hl]            ; [R6] = [R6] - 1

        ld              l, [hl]
        inc             l               ; l <- LOW byte of [R6]

        ld              a, [hl-]        ; a <- LOW byte of [[R6] - 1]
        inc             h
        ld              l, [hl]         ; l <- HIGH byte of [[R6] - 1]

        ld              h, [hl]         ; h <- HIGH of JumpAddr (Real) >> 1
        ld              l, a

        add             hl, hl

        ld              sp, hl          ; sp = JumpAddr (Real)

        RETURN          0               ; return, enabling interrupts

        ; ^ 22 + 4 = 26 cycles vs. 39 cycles (saves 13 cycles)

; ---------------------------------------------------------------------------
;        .::** Code for 16-bit, 10-bit, and 8-bit MVII $xxxx, RD **::.
; ---------------------------------------------------------------------------

ifMVII_16::                             ; MoVe In Immediate (16-bit)

        ; OBJ_FF:       ifMVII_16
        ; OBJ_00:       AA .SHL. 8 + BB
        ; OBJ_01:       R0 + RD .SHL. 1
        ; OBJ_02:       BB .SHL. 8

        DOIMMEDIATE_16                  ; da, de <- AABB, hl -> RD

        ld              [hl+], a        ; store out BB to LOW byte of [RD]
        ld              [hl], d         ; store out AA to HIGH byte of [RD]

        RETURN          0               ; return, enabling interrupts

ifMVII_10::                             ; MoVe In Immediate (10-bit)

        ; OBJ_00:       RD_GTA|ifMVII_10
        ; OBJ_01:       AA .SHL. 8 + BB

        RSING_TAB       l, 6, 0, rtNone ; l <- LOW byte of RD

        DOIMMEDIATE_10                  ; de <- AABB, hl -> RD

        ld              [hl], e         ; store out BB to LOW byte of [RD]
        inc             l
        ld              [hl], d         ; store out AA to HIGH byte of [RD]

        reti                            ; return, enabling interrupts

        ; ^ 2-5 + 5 + 5 + 4 = 16-19 cycles, 33 bytes

ifMVII_10_R6::                          ; MoVe In Immediate (10-bit) to [R6]

        ; OBJ_00:       RD_GTA|ifMVII_10_R6
        ; OBJ_01:       AA .SHL. 8 + BB

        ld              hl, R6

        di

        pop             de              ; de <- AABB

        ld              a, [StackStart]
        add             e

        ld              [hl+], a        ; store out BB to LOW byte of [R6]
        ld              [hl], d         ; store out AA to HIGH byte of [R6]

        reti                            ; return, enabling interrupts

ifMVII_10_01::                          ; MoVe In Immediate (10-bit)

        ; OBJ_00:       ifMVII_10_01
        ; OBJ_01:       BB .SHL. 8 + RD .SHL. 1

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; l <- LOW byte of OBJ_01 (RD)
                                        ; h <- HIGH byte of OBJ_01 (BB)

        ld              a, h            ; a = BB
        ld              h, c            ; hl -> RD

        ld              [hl+], a        ; store out BB to LOW byte of [RD]
        ld              [hl], 1         ; store out AA to HIGH byte of [RD]

        RETURN          0               ; return, enabling interrupts

        ; ^ 15 cycles, 8 bytes

ifMVII_10_02::                          ; MoVe In Immediate (10-bit)

        ; OBJ_00:       ifMVII_10_02
        ; OBJ_01:       BB .SHL. 8 + RD .SHL. 1

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; l <- LOW byte of OBJ_01 (RD)
                                        ; h <- HIGH byte of OBJ_01 (BB)

        ld              a, h            ; a = BB
        ld              h, c            ; hl -> RD

        ld              [hl+], a        ; store out BB to LOW byte of [RD]
        ld              [hl], 2         ; store out AA to HIGH byte of [RD]

        RETURN          0               ; return, enabling interrupts

        ; ^ 15 cycles, 8 bytes

ifMVII_10_03::                          ; MoVe In Immediate (10-bit)

        ; OBJ_00:       ifMVII_10_03
        ; OBJ_01:       BB .SHL. 8 + RD .SHL. 1

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; l <- LOW byte of OBJ_01 (RD)
                                        ; h <- HIGH byte of OBJ_01 (BB)

        ld              a, h            ; a = BB
        ld              h, c            ; hl -> RD

        ld              [hl+], a        ; store out BB to LOW byte of [RD]
        ld              [hl], 3         ; store out AA to HIGH byte of [RD]

        RETURN          0               ; return, enabling interrupts

        ; ^ 15 cycles, 8 bytes

ifMVII_8::                              ; MoVe In Immediate (8-bit)

        ; OBJ_00:       ifMVII_8
        ; OBJ_01:       BB .SHL. 8 + RD .SHL. 1 + 1

        xor             a               ; set a = 0 (to clear HIGH of [RD])

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; l <- LOW byte of OBJ_1 (RD)
                                        ; h <- HIGH byte of OBJ_2 (BB)

        ld              e, h            ; e = BB
        ld              h, c            ; hl -> RD + 1 ( @ HIGH byte )

        ld              [hl-], a        ; zero HIGH byte of [RD]
        ld              [hl], e         ; store out BB to LOW byte of [RD]

        RETURN          0               ; return, enabling interrupts

        ; ^ 15 cycles, 8 bytes

ifMVII_R7::                             ; MoVe In Immediate to [R7]

        ; NOTE: It is assumed that an SDBD would never occur with this
        ;       instruction, since it would take 4 program words to store
        ;       such an instruction, whereas a J/JSR would take only 3.

        ; OBJ_00:       ifMVII_R7
        ; OBJ_01:       AA .SHL. 8 + BB

        ld              hl, Jump_Op

        di                              ; don't allow VBlankFunc to interrupt

        pop             de              ; de <- OBJ_1 (AABB)

        ld              a, d            ; a = HIGH byte of OBJ_1 (AA)

        jp              [hl]            ; convert addr, switch bank if needed

; ---------------------------------------------------------------------------
;         .::** Code for ADD $XXYY, RD with and without flags **::.
; ---------------------------------------------------------------------------

ifADD_SPLO::                            ; ADD from ScratchPad LOw

        ; OBJ_00:       ifADD_SPLO
        ; OBJ_01:       BB .SHL. 9 + RD .SHL. 1

        ld              d, SCRATCHPAD_RAM >> 8

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; l <- LOW byte of OBJ_01 (RD)
                                        ; h <- HIGH byte of OBJ_01 (BB)

        ld              e, h            ; de -> SCRATCHPAD_RAM
        ld              h, c            ; hl -> RD

        ld              a, [de]         ; a <- 8-bit value from ScratchPad

        ADD_RD_8        0               ; [RD] <- [RD] + a

ifADD_SPHI::                            ; ADD from ScratchPad HIgh

        ; OBJ_00:       ifADD_SPHI
        ; OBJ_01:       BB .SHL. 9 + RD .SHL. 1

        ld              d, ( SCRATCHPAD_RAM + $0080 * 2 ) >> 8

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; l <- LOW byte of OBJ_01 (RD)
                                        ; h <- HIGH byte of OBJ_01 (BB)

        ld              e, h            ; de -> SCRATCHPAD_RAM
        ld              h, c            ; hl -> RD

        ld              a, [de]         ; a <- 8-bit value from ScratchPad

        ADD_RD_8        0               ; [RD] <- [RD] + a

ifADD_SPLO_F::                          ; ADD from ScratchPad LOw w/ Flags

        ; OBJ_00:       ifADD_SPLO_F
        ; OBJ_01:       BB .SHL. 9 + RD .SHL. 1

        ld              d, SCRATCHPAD_RAM >> 8

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; l <- LOW byte of OBJ_01 (RD)
                                        ; h <- HIGH byte of OBJ_01 (BB)

        ld              e, h            ; de -> SCRATCHPAD_RAM
        ld              h, c            ; hl -> RD

        ld              a, [de]         ; a <- 8-bit value from ScratchPad

        ADD_RD_8_F      0               ; [RD] <- [RD] + a

ifADD_SPHI_F::                          ; ADD from ScratchPad HIgh w/ Flags

        ; OBJ_00:       ifADD_SPHI_F
        ; OBJ_01:       BB .SHL. 9 + RD .SHL. 1

        ld              d, ( SCRATCHPAD_RAM + $0080 * 2 ) >> 8

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; l <- LOW byte of OBJ_01 (RD)
                                        ; h <- HIGH byte of OBJ_01 (BB)

        ld              e, h            ; de -> SCRATCHPAD_RAM
        ld              h, c            ; hl -> RD

        ld              a, [de]         ; a <- 8-bit value from ScratchPad

        ADD_RD_8_F      0               ; [RD] <- [RD] + a

ifADD_F::                               ; ADD (from any region)

        ; OBJ_00:       RD_GTA|ifADD_F
        ; OBJ_01:       XX .SHL. 8 + YY

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DODIRECT        0               ; da <- [ADDR], hl -> RD
        ADD_RD_F        0               ; [RD] <- [RD] + da, set Status

ifADD_R7::                              ; ADD (from any region) to [R7]

        ; OBJ_00:       ifADD_R7
        ; OBJ_01:       XX .SHL. 8 + YY

        DODIRECT_R7                     ; hl <- ADDR, da <- [R7] (real) >> 1

        add             [hl]            ; a = LOW byte of [R7] + [ADDR]
        ld              e, a            ; save it in e (for Jump_Op)

        inc             l               ; pos @ HIGH byte of [ADDR]

        ldh             a, [R7_Conv]
        adc             d               ; a = HIGH byte of [R7] + CY
        add             [hl]            ;   = HIGH byte of [R7] + [ADDR]

        jp              Jump_Op         ; convert addr, switch bank if needed

        ; ^ 26 cycles, 19 bytes

; ---------------------------------------------------------------------------
;       .::** Code for ADD@ R1-R3, RD with and without flags **::.
; ---------------------------------------------------------------------------

ifADD@_R13::                            ; ADD implied R1-R3, SDBD or NoSDBD

        ; OBJ_00:       RMRD_GTA|ifADD@_R13

        RPAIR_HEAD      hl, 3, 1, 5, 0, rtRM_P1  ; l <- LOW byte of RM + 1
                                                 ; h <- LOW byte of RD

        DOIMPLIED_R13                   ; da <- [[RM]], hl -> RD
        ADD_RD          mpROMB0         ; [RD] <- [RD] + da

ifADD@_R13_F::                          ; ADD implied R1-R3, SDBD or NoSDBD

        ; OBJ_00:       RMRD_GTA|ifADD@_R13_F

        RPAIR_HEAD      hl, 3, 1, 5, 0, rtRM_P1  ; l <- LOW byte of RM + 1
                                                 ; h <- LOW byte of RD

        DOIMPLIED_R13                   ; da <- [[RM]], hl -> RD
        ADD_RD_F        mpROMB0         ; [RD] <- [RD] + da

ifADD@_R13_R7::                         ; ADD implied R1-R3 to [R7]

        ; OBJ_00:       RM_GTA|ifADD@_R13_R7

        RSING_TAB       l, 3, 1, rtRM_P1  ; l <- LOW byte of RM + 1

        DOIMPLIED_R13_R7                ; de <- [[RM]], hl = [R7] (real) >> 1

        add             l               ; a = LOW byte of [R7] + [[RM]]
        ld              e, a            ; save it in e (for Jump_Op)

        ldh             a, [R7_Conv]
        adc             h               ; a = HIGH byte of [R7]
        add             d               ;   = HIGH byte of [R7] + [[RM]]

        jp              Jump_Op         ; convert addr, switch bank if needed

; ---------------------------------------------------------------------------
;       .::** Code for ADD@ R4-R5, RD with and without flags **::.
; ---------------------------------------------------------------------------

ifADD@_R45::                            ; ADD implied R4-R5, SDBD or NoSDBD

        ; OBJ_00:       RMRD_GTA|ifADD@_R45

        RPAIR_HEAD      hl, 5, 4, 5, 0, rtNone  ; l <- LOW byte of RM
                                                ; h <- LOW byte of RD

        DOIMPLIED_R45                   ; da <- [[RM]], hl -> RD
        ADD_RD          mpROMB0         ; [RD] <- [RD] + da

ifADD@_R45_F::                          ; ADD implied R4-R5, SDBD or NoSDBD

        ; OBJ_00:       RMRD_GTA|ifADD@_R45_F

        RPAIR_HEAD      hl, 5, 4, 5, 0, rtNone  ; l <- LOW byte of RM
                                                ; h <- LOW byte of RD

        DOIMPLIED_R45                   ; da <- [[RM]], hl -> RD
        ADD_RD_F        mpROMB0         ; [RD] <- [RD] + da, set Status

ifADD@_R45_R7::                         ; ADD implied R4-R5 to [R7]

        ; OBJ_00:       RM_GTA|ifADD@_R45_R7

        RSING_TAB       l, 5, 4, rtNone ; l <- LOW byte of RM

        DOIMPLIED_R45_R7                ; da <- [[RM]], hl = [R7] (real)

        srl             h
        rr              l               ; hl = [R7] (real) >> 1

        add             l               ; a = LOW byte of [R7] + [[RM]]
        ld              e, a            ; save it in e (for Jump_Op)

        ldh             a, [R7_Conv]
        adc             h               ; a = HIGH byte of [R7]
        add             d               ;   = HIGH byte of [R7] + [[RM]]

        jp              Jump_Op         ; convert addr, switch bank if needed

; ---------------------------------------------------------------------------
;         .::** Code for ADD@ R6, RD with and without flags **::.
; ---------------------------------------------------------------------------

ifADD@_R6::                             ; ADD implied R6

        ; OBJ_00:       RD_GTA|ifADD@_R6

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DOIMPLIED_R6                    ; da <- [[R6] - 1], hl -> RD
        ADD_RD          0               ; [RD] <- [RD] + da

ifADD@_R6_F::                           ; ADD implied R6 w/ Flags

        ; OBJ_00:       RD_GTA|ifADD@_R6_F

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DOIMPLIED_R6                    ; da <- [[R6] - 1], hl -> RD
        ADD_RD_F        0               ; [RD] <- [RD] + da, set Status

ifADD@_R6_R7::                          ; ADD implied R6 to [R7]

        ; OBJ_00:       ifADD@_R6_R7

        ; $6123 (CP1610) -> $6246 (REAL)

        DOIMPLIED_R6_R7                 ; da <- [[R6] - 1], hl = [R7] (real)

        srl             h
        rr              l               ; hl = [R7] (real) >> 1

        add             l               ; a = LOW byte of [R7] + [[R6] - 1]
        ld              e, a            ; save it in e (for Jump_Op)

        ldh             a, [R7_Conv]
        adc             h
        add             d               ; a = HIGH byte of [R7] + [[R6] - 1]

        res             0, b            ; * reset C (see BEAMRIDER)

        jp              Jump_Op         ; convert addr, switch bank if needed

; ---------------------------------------------------------------------------
;      .::** Code for 16-bit, 10-bit, and 8-bit ADDI $XXYY, RD **::.
; ---------------------------------------------------------------------------

ifADDI_16::                             ; ADD Immediate (16-bit)

        ; OBJ_FF:       ifADDI_16
        ; OBJ_00:       AA .SHL. 8 + BB
        ; OBJ_01:       R0 + RD .SHL. 1
        ; OBJ_02:       BB .SHL. 8

        DOIMMEDIATE_16                  ; da, de <- AABB, hl -> RD
        ADD_RD          0               ; [RD] <- [RD] + da

ifADDI_16_F::                           ; ADD Immediate (16-bit) w/ Flags

        ; OBJ_FF:       ifADDI_16_F
        ; OBJ_00:       AA .SHL. 8 + BB
        ; OBJ_01:       R0 + RD .SHL. 1
        ; OBJ_02:       BB .SHL. 8

        DOIMMEDIATE_16                  ; da, de <- AABB, hl -> RD
        ADD_RD_F        0               ; [RD] <- [RD] + da, set Status

ifADDI_10::                             ; ADD Immediate (10-bit)

        ; OBJ_00:       RD_GTA|ifADDI_10
        ; OBJ_01:       AA .SHL. 8 + BB

        RSING_TAB       l, 6, 0, rtNone ; l <- LOW byte of RD

        DOIMMEDIATE_10                  ; de <- AABB, hl -> RD
        ld              a, e            ; a = BB
        ADD_RD          0               ; [RD] <- [RD] + da

ifADDI_10_F::                           ; ADD Immediate (10-bit) w/ Flags

        ; OBJ_00:       RD_GTA|ifADDI_10_F
        ; OBJ_01:       AA .SHL. 8 + BB

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        DOIMMEDIATE_10                  ; de <- AABB, hl -> RD
        ld              a, e            ; a = BB
        ADD_RD_F        0               ; [RD] <- [RD] + da, set Status

ifADDI_8::                              ; ADD Immediate (8-bit)

        ; OBJ_00:       ifADDI_8
        ; OBJ_01:       BB .SHL. 8 + RD .SHL. 1

        DOIMMEDIATE_8                   ; h <- BB, l -> RD
        ld              a, h            ; a = BB
        ld              h, c            ; hl -> RD
        ADD_RD_8        0               ; [RD] <- [RD] + a

ifADDI_8_F::                            ; ADD Immediate (8-bit) w/ Flags

        ; OBJ_00:       ifADDI_8_F
        ; OBJ_01:       BB .SHL. 8 + RD .SHL. 1

        DOIMMEDIATE_8                   ; h <- BB, l -> RD
        ld              a, h            ; a = BB
        ld              h, c            ; hl -> RD
        ADD_RD_8_F      0               ; [RD] <- [RD] + a, set Status

ifADDI_R7::                             ; ADD Immediate to [R7]

        ; NOTE: Although there is already a perfectly good 2-decle
        ;       instruction that serves this function (see ifBcond), this
        ;       seems to be occassionally used anyway (see BurgerTime or
        ;       Shark Shark!).  It is still assumed, however, that it
        ;       won't occur w/ SDBD, since it would take as many as 4
        ;       program words to store (a jump instruction would take 3).

        ; OBJ_00:       ifB
        ; OBJ_01:       ( SP + 2 * 2 ) + ( AA .SHL. 8 + BB ) * 2

; ---------------------------------------------------------------------------
;          .::** Code for SUB $XXYY, RD (with flags only) **::.
; ---------------------------------------------------------------------------

ifSUB_F::                               ; SUBtract (from any region) w/ Flags

        ; OBJ_00:       RD_GTA|ifSUB_F
        ; OBJ_01:       XX .SHL. 8 + YY

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DODIRECT        mpSub           ; da <- [ADDR], hl -> RD
        ld              e, a            ; e = LOW byte of [ADDR]
        SUB_RD_F        mpSub           ; [RD] <- [RD] - de, set Status

ifSUB_R7::                              ; SUBtract (from any region) to [R7]

        ; OBJ_00:       ifSUB_R7
        ; OBJ_01:       XX .SHL. 8 + YY

        DODIRECT_R7                     ; hl <- ADDR, da = [R7] (real) >> 1

        sub             [hl]            ; a = LOW byte of [R7] - [ADDR]
        ld              e, a            ; save it in e (for Jump_Op)

        inc             l               ; pos @ HIGH byte of [ADDR]

        ldh             a, [R7_Conv]
        sbc             [hl]
        add             d               ; a = HIGH byte of [R7] - [ADDR]

        jp              Jump_Op         ; convert addr, switch bank if needed

; ---------------------------------------------------------------------------
;       .::** Code for SUB@ R1-R3, RD with and without flags **::.
; ---------------------------------------------------------------------------

ifSUB@_R13::                            ; SUBtract implied R1-R3

        ; OBJ_00:       RMRD_GTA|ifSUB@_R13

        RPAIR_HEAD      hl, 3, 1, 5, 0, rtRM_P1  ; l <- LOW byte of RM + 1
                                                 ; h <- LOW byte of RD

        DOIMPLIED_R13                   ; da <- [[RM]], hl -> RD
        ld              e, a
        SUB_RD          mpROMB0         ; [RD] <- [RD] - de

ifSUB@_R13_F::                          ; SUBtract implied R1-R3 w/ Flags

        ; OBJ_00:       RMRD_GTA|ifSUB@_R13_F

        RPAIR_HEAD      hl, 3, 1, 5, 0, rtRM_P1  ; l <- LOW byte of RM + 1
                                                 ; h <- LOW byte of RD

        DOIMPLIED_R13                   ; da <- [[RM]], hl -> RD
        ld              e, a
        SUB_RD_F        mpROMB0 | mpSub ; [RD] <- [RD] - de, set Status

ifSUB@_R13_R7::                         ; SUBtract implied R1-R3 from [R7]

        ; OBJ_00:       RM_GTA|ifSUB@_R13_R7

        RSING_TAB       l, 3, 1, rtRM_P1  ; l <- LOW byte of RM + 1

        DOIMPLIED_R13_R7                ; da <- [[RM]], hl = [R7] (CP1610)

        ld              e, a            ; e = LOW byte of [[RM]]

        ld              a, l            ; a = LOW byte of [R7]
        sub             e               ;   = LOW byte of [R7] - [[RM]]
        ld              e, a            ; save it in e (for Jump_Op)

        ldh             a, [R7_Conv]
        sbc             d
        add             h               ; a = HIGH byte of [R7] - [[RM]]

        jp              Jump_Op         ; convert addr, switch bank if needed

; ---------------------------------------------------------------------------
;       .::** Code for SUB@ R4-R5, RD with and without flags **::.
; ---------------------------------------------------------------------------

ifSUB@_R45::                            ; SUBtract implied R4-R5

        ; OBJ_00:       RMRD_GTA|ifSUB@_R45

        RPAIR_HEAD      hl, 5, 4, 5, 0, rtNone  ; l <- LOW byte of RM
                                                ; h <- LOW byte of RD

        DOIMPLIED_R45                   ; da <- [[RM]], hl -> RD
        ld              e, a
        SUB_RD          mpROMB0         ; [RD] <- [RD] - de

ifSUB@_R45_F::                          ; SUBtract implied R4-R5 w/ Flags

        ; OBJ_00:       RMRD_GTA|ifSUB@_R45_F

        RPAIR_HEAD      hl, 5, 4, 5, 0, rtNone  ; l <- LOW byte of RM
                                                ; h <- LOW byte of RD

        DOIMPLIED_R45                   ; da <- [[RM]], hl -> RD
        ld              e, a
        SUB_RD_F        mpROMB0 | mpSub ; [RD] <- [RD] - de, set Status

ifSUB@_R45_R7::                         ; SUBtract implied R4-R5 from [R7]

        ; OBJ_00:       RM_GTA|ifSUB@_R45_R7

        RSING_TAB       l, 5, 4, rtNone ; l <- LOW byte of RM

        DOIMPLIED_R45_R7                ; da <- [[RM]], hl = [R7] (real)

        ld              e, a            ; e = LOW byte of [[RM]]

        srl             h
        ld              a, l
        rra                             ; ha = [R7] (real) >> 1

        sub             e               ; a = LOW byte of [R7] - [[RM]]
        ld              e, a            ; save it in e (for Jump_Op)

        ldh             a, [R7_Conv]
        sbc             d
        add             h               ; a = HIGH byte of [R7] - [[RM]]

        jp              Jump_Op         ; convert addr, switch bank if needed

; ---------------------------------------------------------------------------
;         .::** Code for SUB@ R6, RD with and without flags **::.
; ---------------------------------------------------------------------------

ifSUB@_R6::                             ; SUBtract implied R6

        ; OBJ_00:       RD_GTA|ifSUB@_R6

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DOIMPLIED_R6                    ; da <- [[R6] - 1], hl -> RD
        ld              e, a
        SUB_RD          0               ; [RD] <- [RD] - de

ifSUB@_R6_F::                           ; SUBtract implied R6 w/ Flags

        ; OBJ_00:       RD_GTA|ifSUB@_R6_F

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DOIMPLIED_R6                    ; da <- [[R6] - 1], hl -> RD
        ld              e, a
        SUB_RD_F        mpSub           ; [RD] <- [RD] - de, set Status

ifSUB@_R6_R7::                          ; SUBtract implied R6 from [R7]

        ; OBJ_00:       ifSUB@_R6_R7

        DOIMPLIED_R6_R7                 ; da <- [[R6] - 1], hl = [R7] (real)

        ld              e, a

        srl             h
        ld              a, l
        rra                             ; ha = [R7] (real) >> 1

        sub             e               ; a = LOW byte of [R7] - [[R6] - 1]
        ld              e, a            ; save it in e (for Jump_Op)

        ldh             a, [R7_Conv]
        sbc             d
        add             h               ; a = HIGH byte of [R7] - [[R6] - 1]

        jp              Jump_Op         ; convert addr, switch bank if needed

; ---------------------------------------------------------------------------
;      .::** Code for 16-bit, 10-bit, and 8-bit SUBI $XXYY, RD **::.
; ---------------------------------------------------------------------------

ifSUBI_16::                             ; SUBtract Immediate (16-bit)

        ; OBJ_FF:       ifSUBI_16
        ; OBJ_00:       AA .SHL. 8 + BB
        ; OBJ_01:       R0 + RD .SHL. 1
        ; OBJ_02:       BB .SHL. 8

        DOIMMEDIATE_16                  ; da, de <- XXYY, hl -> RD
        SUB_RD          0               ; [RD] <- [RD] - de

ifSUBI_16_F::                           ; SUBtract Immediate (16-bit)

        ; OBJ_FF:       ifSUBI_16_F
        ; OBJ_00:       AA .SHL. 8 + BB
        ; OBJ_01:       R0 + RD .SHL. 1
        ; OBJ_02:       BB .SHL. 8

        DOIMMEDIATE_16                  ; da, de <- XXYY, hl -> RD
        SUB_RD_F        mpSub           ; [RD] <- [RD] - de, set Status

ifSUBI_10::                             ; SUBtract Immediate (10-bit)

        ; OBJ_00:       RD_GTA|ifSUBI_10
        ; OBJ_01:       AA .SHL. 8 + BB

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        DOIMMEDIATE_10                  ; de <- XXYY, hl -> RD
        SUB_RD          0               ; [RD] <- [RD] - de

ifSUBI_10_F::                           ; SUBtract Immediate (10-bit) w/ Flags

        ; OBJ_00:       RD_GTA|ifSUBI_10_F
        ; OBJ_01:       AA .SHL. 8 + BB

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        DOIMMEDIATE_10                  ; de <- XXYY, hl -> RD
        SUB_RD_F        mpSub           ; [RD] <- [RD] - de, set Status

ifSUBI_8::                              ; SUBtract Immediate (8-bit)

        ; OBJ_00:       ifSUBI_8
        ; OBJ_01:       BB .SHL. 8 + RD .SHL. 1

        DOIMMEDIATE_8                   ; h <- BB, l -> RD
        ld              e, h            ; e = BB
        ld              h, c            ; hl -> RD
        SUB_RD_8        0               ; [RD] <- [RD] - e

ifSUBI_8_F::                            ; SUBtract Immediate (8-bit) w/ Flags

        ; OBJ_00:       ifSUBI_8_F
        ; OBJ_01:       BB .SHL. 8 + RD .SHL. 1

        DOIMMEDIATE_8                   ; h <- BB, l -> RD
        ld              e, h            ; e = BB
        ld              h, c            ; hl -> RD
        SUB_RD_8_F      mpSub           ; [RD] <- [RD] - e, set Status

ifSUBI_R7::                             ; SUBtract Immediate from [R7]

        ; OBJ_00:       ifB
        ; OBJ_01:       ( SP + 2 * 2 ) - ( AA .SHL. 8 + BB ) * 2

; ---------------------------------------------------------------------------
;       .::** CMP direct, implied, and immediate instructions **::.
; ---------------------------------------------------------------------------

ifCMP_SPLO_F::                          ; CoMPare ScratchPad LOw w/ Flags

        ; OBJ_00:       ifCMP_SPLO_F
        ; OBJ_01:       BB .SHL. 9 + RD .SHL. 1

        ld              d, SCRATCHPAD_RAM >> 8

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; l <- LOW byte of OBJ_1 (RD)
                                        ; h <- HIGH byte of OBJ_1 (BB)

        ld              e, h            ; de -> SCRATCHPAD_RAM
        ld              h, c            ; hl -> RD

        ld              a, [de]
        ld              e, a            ; e <- 8-bit value from ScratchPad

        SUB_RD_8_F      mpCmp           ; [RD] - e, set Status

ifCMP_SPHI_F::                          ; CoMPare ScratchPad HIgh w/ Flags

        ; OBJ_00:       ifCMP_SPHI_F
        ; OBJ_01:       BB .SHL. 9 + RD .SHL. 1

        ld              d, ( SCRATCHPAD_RAM + $080 * 2 ) >> 8

        di                              ; don't allow VBlankFunc to interrupt

        pop             hl              ; l <- LOW byte of OBJ_1 (RD)
                                        ; h <- HIGH byte of OBJ_1 (BB)

        ld              e, h            ; de -> SCRATCHPAD_RAM
        ld              h, c            ; hl -> RD

        ld              a, [de]
        ld              e, a            ; e <- 8-bit value from ScratchPad

        SUB_RD_8_F      mpCmp           ; [RD] - e, set Status

ifCMP_F::                               ; CoMPare (to any region) w/ Flags

        ; OBJ_00:       RD_GTA|ifCMP_F
        ; OBJ_01:       XX .SHL. 8 + YY

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DODIRECT        mpCmp           ; da <- [ADDR], hl -> RD
        ld              e, a            ; e = LOW byte of [ADDR]
        SUB_RD_F        mpCmp           ; [RD] - de, set Status

; ---------------------------------------------------------------------------
;         .::** Code for CMP@ R1-R6, RD (with flags only) **::.
; ---------------------------------------------------------------------------

ifCMP@_R13_F::                          ; CoMPare implied R1-R3, with Flags

        ; OBJ_00:       RMRD_GTA|ifCMP@_R13_F

        RPAIR_HEAD      hl, 3, 1, 5, 0, rtRM_P1  ; l <- LOW byte of RM + 1
                                                 ; h <- LOW byte of RD

        DOIMPLIED_R13                   ; da <- [[RM]], hl -> RD
        ld              e, a
        SUB_RD_F        mpROMB0 | mpCmp ; [RD] - de, set Status

ifCMP@_R45_F::                          ; CoMPare implied R4-R5, with Flags

        ; OBJ_00:       RMRD_GTA|ifCMP@_R45_F

        RPAIR_HEAD      hl, 5, 4, 5, 0, rtNone  ; l <- LOW byte of RM
                                                ; h <- LOW byte of RD

        DOIMPLIED_R45                   ; da <- [[RM]], hl -> RD
        ld              e, a
        SUB_RD_F        mpROMB0 | mpCmp ; [RD] - de, set Status

ifCMP@_R6_F::                           ; CoMPare implied R6, with Flags

        ; OBJ_00:       RD_GTA|ifCMP@_R6_F

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DOIMPLIED_R6                    ; da <- [[R6] - 1], hl -> RD
        ld              e, a
        SUB_RD_F        mpCmp           ; [RD] - de, set Status

; ---------------------------------------------------------------------------
;        .::** Code for 16-bit, 10-bit, and 8-bit CMPI $xxxx, RD **::.
; ---------------------------------------------------------------------------

ifCMPI_16_F::                           ; CoMPare Immediate (16-bit) w/ Flags

        ; OBJ_FF:       ifCMPI_16_F
        ; OBJ_00:       AA .SHL. 8 + BB
        ; OBJ_01:       R0 + RD .SHL. 1
        ; OBJ_02:       BB .SHL. 8

        DOIMMEDIATE_16                  ; da, de <- XXYY, hl -> RD
        SUB_RD_F        mpCmp           ; [RD] - de, set Status

ifCMPI_10_F::                           ; CoMPare Immediate (10-bit) w/ Flags

        ; OBJ_00:       RD_GTA|ifCMPI_10_F
        ; OBJ_01:       AA .SHL. 8 + BB

        RSING_TAB       l, 6, 0, rtNone ; l <- LOW byte of RD

        DOIMMEDIATE_10                  ; de <- XXYY, hl -> RD
        SUB_RD_F        mpCmp           ; [RD] - de, set Status

ifCMPI_8_F::                            ; CoMPare Immediate (8-bit) w/ Flags

        ; OBJ_00:       ifCMPI_8_F
        ; OBJ_01:       BB .SHL. 8 + RD .SHL. 1

        DOIMMEDIATE_8                   ; h <- BB, hl -> RD
        ld              e, h            ; e = BB
        ld              h, c            ; hl -> RD
        SUB_RD_8_F      mpCmp           ; [RD] - e, set Status

; ---------------------------------------------------------------------------
;         .::** AND direct, implied, and immediate instructions **::.
; ---------------------------------------------------------------------------

ifAND_F::                               ; AND (with any region) w/ Flags

        ; OBJ_00:       RD_GTA|ifAND_F
        ; OBJ_01:       XX .SHL. 8 + YY

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DODIRECT        0               ; da <- [ADDR], hl -> RD
        AND_RD_F        0               ; [RD] <- [RD] & da, set Status

ifAND@_R13::                            ; AND implied R1-R3

        ; OBJ_00:       RMRD_GTA|ifAND@_R13

        RPAIR_HEAD      hl, 3, 1, 5, 0, rtRM_P1  ; l <- LOW byte of RM + 1
                                                 ; h <- LOW byte of RD

        DOIMPLIED_R13                   ; da <- [[RM]], hl -> RD
        AND_RD          mpROMB0         ; [RD] <- [RD] & da

ifAND@_R13_F::                          ; AND implied R1-R3 w/ Flags

        ; OBJ_00:       RMRD_GTA|ifAND@_R13_F

        RPAIR_HEAD      hl, 3, 1, 5, 0, rtRM_P1  ; l <- LOW byte of RM + 1
                                                 ; h <- LOW byte of RD

        DOIMPLIED_R13                   ; da <- [[RM]], hl -> RD
        AND_RD_F        mpROMB0         ; [RD] <- [RD] & da, set Status

ifAND@_R45::                            ; AND implied R4-R5

        ; OBJ_00:       RMRD_GTA|ifAND@_R45

        RPAIR_HEAD      hl, 5, 4, 5, 0, rtNone  ; l <- LOW byte of RM
                                                ; h <- LOW byte of RD

        DOIMPLIED_R45                   ; da <- [[RM]], hl -> RD
        AND_RD          mpROMB0         ; [RD] <- [RD] & da

ifAND@_R45_F::                          ; AND implied R4-R5 w/ Flags

        ; OBJ_00:       RMRD_GTA|ifAND@_R45_F

        RPAIR_HEAD      hl, 5, 4, 5, 0, rtNone  ; l <- LOW byte of RM
                                                ; h <- LOW byte of RD

        DOIMPLIED_R45                   ; da <- [[RM]], hl -> RD
        AND_RD_F        mpROMB0         ; [RD] <- [RD] & da, set Status

ifAND@_R6::                             ; AND implied R6

        ; OBJ_00:       RD_GTA|ifAND@_R6

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DOIMPLIED_R6                    ; da <- [[R6] - 1], hl -> RD
        AND_RD          0               ; [RD] <- [RD] & da

ifAND@_R6_F::                           ; AND implied R6 w/ Flags

        ; OBJ_00:       RD_GTA|ifAND@_R6_F

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DOIMPLIED_R6                    ; da <- [[R6] - 1], hl -> RD
        AND_RD_F        0               ; [RD] <- [RD] & da, set Status

; ---------------------------------------------------------------------------
;        .::** Code for 16-bit, 10-bit, and 8-bit ANDI $xxxx, RD **::.
; ---------------------------------------------------------------------------

ifANDI_16::                             ; AND Immediate (16-bit)

        ; OBJ_FF:       ifANDI_16
        ; OBJ_00:       AA .SHL. 8 + BB
        ; OBJ_01:       R0 + RD .SHL. 1
        ; OBJ_02:       BB .SHL. 8

        DOIMMEDIATE_16                  ; da, de <- AABB, hl -> RD
        AND_RD          0               ; [RD] <- [RD] & da

ifANDI_16_F::                           ; AND Immediate (16-bit) w/ Flags

        ; OBJ_FF:       ifANDI_16_F
        ; OBJ_00:       AA .SHL. 8 + BB
        ; OBJ_01:       R0 + RD .SHL. 1
        ; OBJ_02:       BB .SHL. 8

        DOIMMEDIATE_16                  ; da, de <- AABB, hl -> RD
        AND_RD_F        0               ; [RD] <- [RD] & da, set Status

ifANDI_10::                             ; AND Immediate (10-bit)

        ; OBJ_00:       RD_GTA|ifANDI_10
        ; OBJ_01:       AA .SHL. 8 + BB

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        DOIMMEDIATE_10                  ; de <- AABB, hl -> RD
        ld              a, e            ; a = BB
        AND_RD          0               ; [RD] <- [RD] & da

ifANDI_10_F::                           ; AND Immediate (10-bit) w/ Flags

        ; OBJ_00:       RD_GTA|ifANDI_10_F
        ; OBJ_01:       AA .SHL. 8 + BB

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        DOIMMEDIATE_10                  ; de <- AABB, hl -> RD
        ld              a, e            ; a = BB
        AND_RD_F        0               ; [RD] <- [RD] & da, set Status

ifANDI_8::                              ; AND Immediate (8-bit)

        ; OBJ_00:       ifANDI_8
        ; OBJ_01:       BB .SHL. 8 + RD .SHL. 1

        DOIMMEDIATE_8                   ; h <- BB, l -> RD
        ld              a, h            ; a = BB
        ld              h, c            ; hl -> RD
        AND_RD_8        0               ; [RD] <- [RD] & a

ifANDI_8_F::                            ; AND Immediate (8-bit) w/ Flags

        ; OBJ_00:       ifANDI_8_F
        ; OBJ_01:       BB .SHL. 8 + RD .SHL. 1

        DOIMMEDIATE_8                   ; h <- BB, l -> RD
        ld              a, h            ; a = BB
        ld              h, c            ; hl -> RD
        AND_RD_8_F      0               ; [RD] <- [RD] & a, set Status

; ---------------------------------------------------------------------------
;         .::** XOR direct, implied, and immediate instructions **::.
; ---------------------------------------------------------------------------

ifXOR_F::                               ; XOR (with any region) w/ Flags

        ; OBJ_00:       RD_GTA|ifXOR_F
        ; OBJ_01:       XX .SHL. 8 + YY

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DODIRECT        0               ; da <- [ADDR], hl -> RD
        XOR_RD_F        0               ; [RD] <- [RD] ^ da, set Status

ifXOR@_R13::                            ; XOR implied R1-R3

        ; OBJ_00:       RMRD_GTA|ifXOR@_R13

        RPAIR_HEAD      hl, 3, 1, 5, 0, rtRM_P1  ; l <- LOW byte of RM + 1
                                                 ; h <- LOW byte of RD

        DOIMPLIED_R13                   ; da <- [[RM]], hl -> RD
        XOR_RD          mpROMB0         ; [RD] <- [RD] ^ da

ifXOR@_R13_F::                          ; XOR implied R1-R3 w/ Flags

        ; OBJ_00:       RMRD_GTA|ifXOR@_R13_F

        RPAIR_HEAD      hl, 3, 1, 5, 0, rtRM_P1  ; l <- LOW byte of RM + 1
                                                 ; h <- LOW byte of RD

        DOIMPLIED_R13                   ; da <- [[RM]], hl -> RD
        XOR_RD_F        mpROMB0         ; [RD] <- [RD] ^ da, set Status

ifXOR@_R45::                            ; XOR implied R4-R5

        ; OBJ_00:       RMRD_GTA|ifXOR@_R45

        RPAIR_HEAD      hl, 5, 4, 5, 0, rtNone  ; l <- LOW byte of RM
                                                ; h <- LOW byte of RD

        DOIMPLIED_R45                   ; da <- [[RM]], hl -> RD
        XOR_RD          mpROMB0         ; [RD] <- [RD] ^ da

ifXOR@_R45_F::                          ; XOR implied R4-R5 w/ Flags

        ; OBJ_00:       RMRD_GTA|ifXOR@_R45_F

        RPAIR_HEAD      hl, 5, 4, 5, 0, rtNone  ; l <- LOW byte of RM
                                                ; h <- LOW byte of RD

        DOIMPLIED_R45                   ; da <- [[RM]], hl -> RD
        XOR_RD_F        mpROMB0         ; [RD] <- [RD] ^ da, set Status

ifXOR@_R6::                             ; XOR implied R6

        ; OBJ_00:       RD_GTA|ifXOR@_R6

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DOIMPLIED_R6                    ; da <- [[R6] - 1], hl -> RD
        XOR_RD          0               ; [RD] <- [RD] ^ da

ifXOR@_R6_F::                           ; XOR implied R6 w/ Flags

        ; OBJ_00:       RD_GTA|ifXOR@_R6_F

        RSING_TAB       e, 5, 0, rtNone ; e <- LOW byte of RD

        DOIMPLIED_R6                    ; da <- [[R6] - 1], hl -> RD
        XOR_RD_F        0               ; [RD] <- [RD] ^ da, set Status

; ---------------------------------------------------------------------------
;        .::** Code for 16-bit, 10-bit, and 8-bit XORI $xxxx, RD **::.
; ---------------------------------------------------------------------------

ifXORI_16::                             ; XOR Immediate (16-bit)

        ; OBJ_FF:       ifXORI_16
        ; OBJ_00:       AA .SHL. 8 + BB
        ; OBJ_01:       R0 + RD .SHL. 1
        ; OBJ_02:       BB .SHL. 8

        DOIMMEDIATE_16                  ; da, de <- AABB, hl -> RD
        XOR_RD          0               ; [RD] <- [RD] ^ da

ifXORI_16_F::                           ; XOR Immediate (16-bit) w/ Flags

        ; OBJ_FF:       ifXORI_16_F
        ; OBJ_00:       AA .SHL. 8 + BB
        ; OBJ_01:       R0 + RD .SHL. 1
        ; OBJ_02:       BB .SHL. 8

        DOIMMEDIATE_16                  ; da, de <- AABB, hl -> RD
        XOR_RD_F        0               ; [RD] <- [RD] ^ da, set Status

ifXORI_10::                             ; XOR Immediate (10-bit)

        ; OBJ_00:       RD_GTA|ifXORI_10
        ; OBJ_01:       AA .SHL. 8 + BB

        RSING_TAB       l, 6, 0, rtNone ; l <- LOW byte of RD

        DOIMMEDIATE_10                  ; de <- AABB, hl -> RD
        ld              a, e            ; a = BB
        XOR_RD          0               ; [RD] <- [RD] ^ da

ifXORI_10_F::                           ; XOR Immediate (10-bit) w/ Flags

        ; OBJ_00:       RD_GTA|ifXORI_10_F
        ; OBJ_01:       AA .SHL. 8 + BB

        RSING_TAB       l, 5, 0, rtNone ; l <- LOW byte of RD

        DOIMMEDIATE_10                  ; de <- AABB, hl -> RD
        ld              a, e            ; a = BB
        XOR_RD_F        0               ; [RD] <- [RD] ^ da, set Status

ifXORI_8::                              ; XOR Immediate (8-bit)

        ; OBJ_00:       ifXORI_8
        ; OBJ_01:       BB .SHL. 8 + RD .SHL. 1

        DOIMMEDIATE_8                   ; h <- BB, l -> RD
        ld              a, h            ; a = BB
        ld              h, c            ; hl -> RD
        XOR_RD_8        0               ; [RD] <- [RD] ^ a

ifXORI_8_F::                            ; XOR Immediate (8-bit) w/ Flags

        ; OBJ_00:       ifXORI_8_F
        ; OBJ_01:       BB .SHL. 8 + RD .SHL. 1

        DOIMMEDIATE_8                   ; h <- BB, l -> RD
        ld              a, h            ; a = BB
        ld              h, c            ; hl -> RD
        XOR_RD_8_F      0               ; [RD] <- [RD] ^ a, set Status

